#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import ctypes
from typing import Any, Dict

bits = 64
psize = bits // 8

dummy_ptr_type = {
	32 : ctypes.c_uint32,
	64 : ctypes.c_uint64
}[bits]

_pointer_type_cache: Dict[str, type] = {}

def PTR(ptype):
	pname = 'c_void' if ptype is None else ptype.__name__

	if pname not in _pointer_type_cache:
		_pointer_type_cache[pname] = type(f'LP_{psize}_{pname}', (dummy_ptr_type,), {})

	return _pointer_type_cache[pname]

VOID = None
INT8  = ctypes.c_byte
INT16 = ctypes.c_int16
INT32 = ctypes.c_int32
INT64 = ctypes.c_int64
INTN  = INT64

UINT8  = ctypes.c_ubyte
UINT16 = ctypes.c_uint16
UINT32 = ctypes.c_uint32
UINT64 = ctypes.c_uint64
UINTN  = UINT64

BOOLEAN = UINT8
CHAR8 = UINT8
CHAR16 = UINT16

FUNCPTR = lambda *args: PTR(ctypes.CFUNCTYPE(*args))
UNION = ctypes.Union
# SIZEOF = lambda t: ctypes.sizeof(t)
# OFFSETOF = lambda cls, fname: getattr(cls, fname).offset

CPU_STACK_ALIGNMENT = 16
PAGE_SIZE = 0x1000

class STRUCT(ctypes.LittleEndianStructure):
	"""An abstract class for C structures.
	"""

	# Structures are packed by default; when needed, padding should be added
	# manually through placeholder fields
	_pack_ = 1

	def __init__(self):
		pass

	def saveTo(self, ql, address: int) -> None:
		"""Store self contents to a specified memory address.
		"""

		data = bytes(self)

		ql.mem.write(address, data)

	@classmethod
	def loadFrom(cls, ql, address: int) -> Any:
		"""Construct an instance of the structure from saved contents.
		"""

		data = bytes(ql.mem.read(address, cls.sizeof()))

		return cls.from_buffer_copy(data)

	@classmethod
	def sizeof(cls) -> int:
		"""Get the C structure size in bytes.
		"""

		return ctypes.sizeof(cls)

	@classmethod
	def offsetof(cls, fname: str) -> int:
		"""Get the offset of a field in the C structure.
		"""

		return getattr(cls, fname).offset

class EnumMeta(type(ctypes.c_int)):
	def __getattr__(self, key):
		return self._members_.index(key)

class ENUM(ctypes.c_int, metaclass=EnumMeta):
	"""An abstract class for continuous C enums.
	"""

	# a list or tuple of names (strings)
	# names will be enumerate by their corresponding index in the list
	_members_ = []

class EnumUCMeta(type(ctypes.c_int)):
	def __getattr__(self, key):
		return self._members_[key]

class ENUM_UC(ctypes.c_int, metaclass=EnumUCMeta):
	"""An abstract class for uncontinuous C enums.
	"""

	# a dictionary of (names : str, value : int) tuples
	# names will be enumerate by their paired value
	_members_ = {}

__all__ = [
	'VOID',
	'INT8',
	'INT16',
	'INT32',
	'INT64',
	'INTN',
	'UINT8',
	'UINT16',
	'UINT32',
	'UINT64',
	'UINTN',
	'BOOLEAN',
	'CHAR8',
	'CHAR16',

	'PTR',
	'FUNCPTR',
	'STRUCT',
	'UNION',
	'ENUM',
	'ENUM_UC',

	'CPU_STACK_ALIGNMENT',
	'PAGE_SIZE'
]


#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import struct, time, os

from shutil import copyfile
from datetime import datetime


from qiling.exception import *
from qiling.os.windows.const import *

from qiling.os.const import *
from qiling.os.windows.fncc import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *
from qiling.os.windows.structs import *

dllname = 'kernel32_dll'

# DWORD GetFileType(
#   HANDLE hFile
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetFileType(ql, address, params):
    hFile = params["hFile"]
    if hFile == STD_INPUT_HANDLE or hFile == STD_OUTPUT_HANDLE or hFile == STD_ERROR_HANDLE:
        ret = FILE_TYPE_CHAR
    else:
        obj = ql.os.handle_manager.get(hFile)
        if obj is None:
            raise QlErrorNotImplemented("API not implemented")
        else:
            # technically is not always a type_char but.. almost
            ret = FILE_TYPE_CHAR
    return ret


# HANDLE FindFirstFileA(
#  LPCSTR             lpFileName,
#  LPWIN32_FIND_DATAA lpFindFileData
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'LPCSTR': 'POINTER'})
def hook_FindFirstFileA(ql, address, params):
    filename = params['lpFileName']
    pointer = params['lpFindFileData']

    if filename == None:
        return INVALID_HANDLE_VALUE
    elif len(filename) >= MAX_PATH:
        return ERROR_INVALID_PARAMETER
    
    target_dir = os.path.join(ql.rootfs, filename.replace("\\", os.sep))
    print('TARGET_DIR = %s' % target_dir)    
    real_path = ql.os.transform_to_real_path(filename)
    # Verify the directory is in ql.rootfs to ensure no path traversal has taken place
    if not os.path.exists(real_path):
        ql.os.last_error = ERROR_FILE_NOT_FOUND
        return INVALID_HANDLE_VALUE

    # Check if path exists
    filesize = 0
    try:
        f = ql.os.fs_mapper.open(real_path, mode="r")
        filesize = os.path.getsize(real_path).to_bytes(8, byteorder="little")
    except FileNotFoundError:
        ql.os.last_error = ERROR_FILE_NOT_FOUND
        return INVALID_HANDLE_VALUE
    
    # Get size of the file
    file_size_low = (int.from_bytes(filesize, "little")) & 0xffffff
    file_size_high = (int.from_bytes(filesize, "little") >> 32)

    # Create a handle for the path
    new_handle = Handle(obj=f)
    ql.os.handle_manager.append(new_handle)

    # Spoof filetime values
    filetime = datetime.now().microsecond.to_bytes(8, byteorder="little")

    find_data = Win32FindData(
                ql, 
                FILE_ATTRIBUTE_NORMAL, 
                filetime, filetime, filetime, 
                file_size_high, file_size_low,
                0, 0, 
                filename,
                0, 0, 0, 0,)
                
    find_data.write(pointer)

    ret = new_handle.id
    return ret

# HANDLE FindFirstFileExA(
#  LPCSTR             lpFileName,
#  FINDEX_INFO_LEVELS fInfoLevelId,
#  FINDEX_SEARCH_OPS  fSearchOp,
#  LPVOID             lpSearchFilter,
#  DWORD              dwAdditionalFlags
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'LPCSTR': 'POINTER'})
def hook_FindFirstFileExA(ql, address, params):
    pass

# HANDLE FindNextFileA(
#  LPCSTR             lpFileName,
#  LPWIN32_FIND_DATAA lpFindFileData
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'LPCSTR': 'POINTER'})
def hook_FindNextFileA(ql, address, params):
    pass


# BOOL FindClose(
#    HANDLE hFindFile
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_FindClose(ql, address, params):
    pass


# BOOL ReadFile(
#   HANDLE       hFile,
#   LPVOID       lpBuffer,
#   DWORD        nNumberOfBytesToRead,
#   LPDWORD      lpNumberOfBytesRead,
#   LPOVERLAPPED lpOverlapped
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_ReadFile(ql, address, params):
    ret = 1
    hFile = params["hFile"]
    lpBuffer = params["lpBuffer"]
    nNumberOfBytesToRead = params["nNumberOfBytesToRead"]
    lpNumberOfBytesRead = params["lpNumberOfBytesRead"]
    lpOverlapped = params["lpOverlapped"]
    if hFile == STD_INPUT_HANDLE:
        if ql.os.automatize_input:
            # TODO maybe insert a good random generation input
            s = (b"A" * (nNumberOfBytesToRead - 1)) + b"\x00"
        else:
            ql.log.debug("Insert input")
            s = ql.os.stdin.read(nNumberOfBytesToRead)
        slen = len(s)
        read_len = slen
        if slen > nNumberOfBytesToRead:
            s = s[:nNumberOfBytesToRead]
            read_len = nNumberOfBytesToRead
        ql.mem.write(lpBuffer, s)
        ql.mem.write(lpNumberOfBytesRead, ql.pack(read_len))
    else:
        f = ql.os.handle_manager.get(hFile).obj
        data = f.read(nNumberOfBytesToRead)
        ql.mem.write(lpBuffer, data)
        ql.mem.write(lpNumberOfBytesRead, ql.pack32(lpNumberOfBytesRead))
    return ret


# BOOL WriteFile(
#   HANDLE       hFile,
#   LPCVOID      lpBuffer,
#   DWORD        nNumberOfBytesToWrite,
#   LPDWORD      lpNumberOfBytesWritten,
#   LPOVERLAPPED lpOverlapped
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={
    "hFile": HANDLE,
    "lpBuffer": POINTER,
    "nNumberOfBytesToWrite": DWORD,
    "lpNumberOfBytesWritten": POINTER,
    "lpOverlapped": POINTER
})
def hook_WriteFile(ql, address, params):
    ret = 1
    hFile = params["hFile"]
    lpBuffer = params["lpBuffer"]
    nNumberOfBytesToWrite = params["nNumberOfBytesToWrite"]
    lpNumberOfBytesWritten = params["lpNumberOfBytesWritten"]
    lpOverlapped = params["lpOverlapped"]
    if hFile == STD_OUTPUT_HANDLE:
        s = ql.mem.read(lpBuffer, nNumberOfBytesToWrite)
        ql.os.stdout.write(s)
        ql.os.string_appearance(s.decode())
        ql.mem.write(lpNumberOfBytesWritten, ql.pack(nNumberOfBytesToWrite))
    else:
        f = ql.os.handle_manager.get(hFile)
        if f is None:
            # Invalid handle
            ql.os.last_error = ERROR_INVALID_HANDLE
            return 0
        else:
            f = f.obj
        buffer = ql.mem.read(lpBuffer, nNumberOfBytesToWrite)
        f.write(bytes(buffer))
        ql.mem.write(lpNumberOfBytesWritten, ql.pack32(nNumberOfBytesToWrite))
    return ret


def _CreateFile(ql, address, params, name):
    ret = INVALID_HANDLE_VALUE
    s_lpFileName = params["lpFileName"]
    dwDesiredAccess = params["dwDesiredAccess"]
    dwShareMode = params["dwShareMode"]
    lpSecurityAttributes = params["lpSecurityAttributes"]
    dwCreationDisposition = params["dwCreationDisposition"]
    dwFlagsAndAttributes = params["dwFlagsAndAttributes"]
    hTemplateFile = params["hTemplateFile"]

    # access mask DesiredAccess
    mode = ""
    if dwDesiredAccess & GENERIC_WRITE:
        mode += "wb"
    else:
        mode += "r"

    try:
        f = ql.os.fs_mapper.open(s_lpFileName, mode)
    except FileNotFoundError:
        ql.os.last_error = ERROR_FILE_NOT_FOUND
        return INVALID_HANDLE_VALUE

    new_handle = Handle(obj=f)
    ql.os.handle_manager.append(new_handle)
    ret = new_handle.id
    return ret


# HANDLE CreateFileA(
#   LPCSTR                lpFileName,
#   DWORD                 dwDesiredAccess,
#   DWORD                 dwShareMode,
#   LPSECURITY_ATTRIBUTES lpSecurityAttributes,
#   DWORD                 dwCreationDisposition,
#   DWORD                 dwFlagsAndAttributes,
#   HANDLE                hTemplateFile
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={
    "lpFileName": STRING,
    "dwDesiredAccess": DWORD,
    "dwShareMode": DWORD,
    "lpSecurityAttributes": POINTER,
    "dwCreationDisposition": DWORD,
    "dwFlagsAndAttributes": DWORD,
    "hTemplateFile": HANDLE
})
def hook_CreateFileA(ql, address, params):
    ret = _CreateFile(ql, address, params, "CreateFileA")
    return ret


# HANDLE CreateFileW(
#   LPCWSTR                lpFileName,
#   DWORD                 dwDesiredAccess,
#   DWORD                 dwShareMode,
#   LPSECURITY_ATTRIBUTES lpSecurityAttributes,
#   DWORD                 dwCreationDisposition,
#   DWORD                 dwFlagsAndAttributes,
#   HANDLE                hTemplateFile
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_CreateFileW(ql, address, params):
    ret = _CreateFile(ql, address, params, "CreateFileW")
    return ret


# DWORD GetTempPathW(
#   DWORD  nBufferLength,
#   LPWSTR lpBuffer
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetTempPathW(ql, address, params):
    temp = (ql.os.windir + "Temp" + "\\\x00").encode('utf-16le')
    dest = params["lpBuffer"]
    temp_path = os.path.join(ql.rootfs, "Windows", "Temp")
    if not os.path.exists(temp_path):
        os.makedirs(temp_path, 0o755)
    ql.mem.write(dest, temp)
    return len(temp)

# DWORD GetTempPathA(
#   DWORD  nBufferLength,
#   LPSTR lpBuffer
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={
    "nBufferLength": DWORD,
    "lpBuffer": POINTER
})
def hook_GetTempPathA(ql, address, params):
    temp = (ql.os.windir + "Temp" + "\\\x00").encode('utf-8')
    dest = params["lpBuffer"]
    temp_path = os.path.join(ql.rootfs, "Windows", "Temp")
    if not os.path.exists(temp_path):
        os.makedirs(temp_path, 0o755)
    ql.mem.write(dest, temp)
    return len(temp)

# DWORD GetShortPathNameW(
#   LPCWSTR lpszLongPath,
#   LPWSTR  lpszShortPath,
#   DWORD   cchBuffer
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetShortPathNameW(ql, address, params):
    paths = params["lpszLongPath"].split("\\")
    dst = params["lpszShortPath"]
    max_size = params["cchBuffer"]
    res = paths[0]
    for path in paths[1:]:
        nameAndExt = path.split(".")
        name = nameAndExt[0]
        ext = "" if len(nameAndExt) == 1 else "." + nameAndExt[1]
        if len(name) > 8:
            name = name[:6] + "~1"
        res += "\\" + name + ext
    res += "\x00"
    res = res.encode("utf-16le")
    if max_size < len(res):
        return len(res)
    else:
        ql.mem.write(dst, res)
    return len(res) - 1


# BOOL GetVolumeInformationW(
#   LPCWSTR lpRootPathName,
#   LPWSTR  lpVolumeNameBuffer,
#   DWORD   nVolumeNameSize,
#   LPDWORD lpVolumeSerialNumber,
#   LPDWORD lpMaximumComponentLength,
#   LPDWORD lpFileSystemFlags,
#   LPWSTR  lpFileSystemNameBuffer,
#   DWORD   nFileSystemNameSize
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetVolumeInformationW(ql, address, params):
    root = params["lpRootPathName"]
    if root != 0:
        pt_volume_name = params["lpVolumeNameBuffer"]
        if pt_volume_name != 0:
            # TODO implement
            volume_name = (ql.os.profile["VOLUME"]["name"] + "\x00").encode("utf-16le")

            ql.mem.write(pt_volume_name, volume_name)

        lpMaximumComponentLength = params["lpMaximumComponentLength"]
        if lpMaximumComponentLength != 0:
            ql.mem.write(lpMaximumComponentLength, (255).to_bytes(2, byteorder="little"))
        pt_serial_number = params["lpVolumeSerialNumber"]
        if pt_serial_number != 0:
            # TODO maybe has to be int
            serial_number = (ql.os.profile["VOLUME"]["serial_number"] + "\x00").encode("utf-16le")
            ql.mem.write(pt_serial_number, serial_number)
        pt_system_type = params["lpFileSystemNameBuffer"]
        pt_flag = params["lpFileSystemFlags"]
        if pt_flag != 0:
            # TODO implement
            flag = 0x00020000.to_bytes(4, byteorder="little")
            ql.mem.write(pt_flag, flag)
        if pt_system_type != 0:
            system_type = (ql.os.profile["VOLUME"]["type"] + "\x00").encode("utf-16le")
            ql.mem.write(pt_system_type, system_type)
    else:
        raise QlErrorNotImplemented("API not implemented")
    return 1


# UINT GetDriveTypeW(
#   LPCWSTR lpRootPathName
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'LPCWSTR': 'POINTER'})
def hook_GetDriveTypeW(ql, address, params):
    path = params["lpRootPathName"]
    if path != 0:
        if path == ql.os.profile["PATH"]["systemdrive"]:
            return DRIVE_FIXED
        # TODO add configuration for drives
    else:
        raise QlErrorNotImplemented("API not implemented")
    return DRIVE_NO_ROOT_DIR


# BOOL GetDiskFreeSpaceW(
#   LPCWSTR lpRootPathName,
#   LPDWORD lpSectorsPerCluster,
#   LPDWORD lpBytesPerSector,
#   LPDWORD lpNumberOfFreeClusters,
#   LPDWORD lpTotalNumberOfClusters
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'LPCWSTR': 'POINTER'})
def hook_GetDiskFreeSpaceW(ql, address, params):
    path = params["lpRootPathName"]
    if path == ql.os.profile["PATH"]["systemdrive"]:
        pt_sectors = params["lpSectorsPerCluster"]
        pt_bytes = params["lpBytesPerSector"]
        pt_free_clust = params["lpNumberOfFreeClusters"]
        pt_total_clust = params["lpTotalNumberOfClusters"]
        sectors = ql.os.profile.getint("VOLUME", "sectors_per_cluster").to_bytes(4, byteorder="little")
        bytes = ql.os.profile.getint("VOLUME", "bytes_per_sector").to_bytes(4, byteorder="little")
        free_clust = ql.os.profile.getint("VOLUME", "number_of_free_clusters").to_bytes(4, byteorder="little")
        total_clust = ql.os.profile.getint("VOLUME", "number_of_clusters").to_bytes(4, byteorder="little")
        ql.mem.write(pt_sectors, sectors)
        ql.mem.write(pt_bytes, bytes)
        ql.mem.write(pt_free_clust, free_clust)
        ql.mem.write(pt_total_clust, total_clust)
    else:
        raise QlErrorNotImplemented("API not implemented")
    return 0


# BOOL CreateDirectoryA(
#  LPCSTR                lpPathName,
#  LPSECURITY_ATTRIBUTES lpSecurityAttributes
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_CreateDirectoryA(ql, address, params):
    path_name = params["lpPathName"]
    target_dir = os.path.join(ql.rootfs, path_name.replace("\\", os.sep))
    print('TARGET_DIR = %s' % target_dir)
    real_path = ql.os.transform_to_real_path(path_name)
    # Verify the directory is in ql.rootfs to ensure no path traversal has taken place
    if not os.path.exists(real_path):
        os.mkdir(real_path)
        return 1
    else:
        ql.os.last_error = ERROR_ALREADY_EXISTS
        return 0


# DWORD GetFileSize(
#  HANDLE  hFile,
#  LPDWORD lpFileSizeHigh
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params_type={'LPDWORD': 'DWORD'})
def hook_GetFileSize(ql, address, params):
    try:
        handle = ql.handle_manager.get(params['hFile'].file)
        return os.path.getsize(handle.name)
    except:
        ql.os.last_error = ERROR_INVALID_HANDLE 
        return 0xFFFFFFFF #INVALID_FILE_SIZE

# HANDLE CreateFileMappingA(
#   HANDLE                hFile,
#   LPSECURITY_ATTRIBUTES lpFileMappingAttributes,
#   DWORD                 flProtect,
#   DWORD                 dwMaximumSizeHigh,
#   DWORD                 dwMaximumSizeLow,
#   LPCSTR                lpName
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={
    "hFile": HANDLE,
    "lpFileMappingAttributes": POINTER,
    "flProtect": DWORD,
    "dwMaximumSizeHigh": DWORD,
    "dwMaximumSizeLow": DWORD,
    "lpName": STRING,
})
def hook_CreateFileMappingA(ql, address, params):
    hFile = params['hFile']
    lpName = params['lpName']
    new_handle = Handle(obj=hFile, name=lpName)
    ql.os.handle_manager.append(new_handle)
    ret = new_handle.id

    return ret

# HANDLE CreateFileMappingW(
#   HANDLE                hFile,
#   LPSECURITY_ATTRIBUTES lpFileMappingAttributes,
#   DWORD                 flProtect,
#   DWORD                 dwMaximumSizeHigh,
#   DWORD                 dwMaximumSizeLow,
#   LPCWSTR               lpName
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={
    "hFile": HANDLE,
    "lpFileMappingAttributes": POINTER,
    "flProtect": DWORD,
    "dwMaximumSizeHigh": DWORD,
    "dwMaximumSizeLow": DWORD,
    "lpName": WSTRING,
})
def hook_CreateFileMappingW(ql, address, params):
    hFile = params['hFile']
    lpName = params['lpName']
    new_handle = Handle(obj=hFile, name=lpName)
    ql.os.handle_manager.append(new_handle)
    ret = new_handle.id

    return ret

# LPVOID MapViewOfFile(
#   HANDLE hFileMappingObject,
#   DWORD  dwDesiredAccess,
#   DWORD  dwFileOffsetHigh,
#   DWORD  dwFileOffsetLow,
#   SIZE_T dwNumberOfBytesToMap
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={
    "hFileMappingObject": HANDLE,
    "dwDesiredAccess": DWORD,
    "dwFileOffsetHigh": DWORD,
    "dwFileOffsetLow": DWORD,
    "dwNumberOfBytesToMap": DWORD
})
def hook_MapViewOfFile(ql, address, params):
    hFileMappingObject = params['hFileMappingObject']
    dwFileOffsetLow = params['dwFileOffsetLow']
    dwNumberOfBytesToMap = params['dwNumberOfBytesToMap']

    map_file_handle = ql.os.handle_manager.search_by_obj(hFileMappingObject)
    if map_file_handle is None:
        ret = ql.os.heap.alloc(dwNumberOfBytesToMap)
        new_handle = Handle(obj=hFileMappingObject, name=ret)
        ql.os.handle_manager.append(new_handle)
    else:
        ret = map_file_handle.name
    hFile = ql.os.handle_manager.get(hFileMappingObject).obj
    if ql.os.handle_manager.get(hFile):
        f = ql.os.handle_manager.get(hFile).obj
        if type(f) is file:
            f.seek(dwFileOffsetLow, 0)
            data = f.read(dwNumberOfBytesToMap)
            ql.mem.write(ret, data)
    return ret


# BOOL UnmapViewOfFile(
#   LPCVOID lpBaseAddress
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={
    "lpBaseAddress": POINTER
})
def hook_UnmapViewOfFile(ql, address, params):
    lpBaseAddress = params['lpBaseAddress']

    map_file_hande = ql.os.handle_manager.search(lpBaseAddress)
    if not map_file_hande:
        return 0
    ql.os.heap.free(map_file_hande.name)
    ql.os.handle_manager.delete(map_file_hande.id)
    return 1



# BOOL CopyFileA(
#   LPCSTR lpExistingFileName,
#   LPCSTR lpNewFileName,
#   BOOL   bFailIfExists
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={
    "lpExistingFileName": STRING,
    "lpNewFileName": STRING,
    "bFailIfExists": DWORD
})
def hook_CopyFileA(ql, address, params):
    lpExistingFileName = ql.os.transform_to_real_path(params["lpExistingFileName"])
    lpNewFileName = ql.os.transform_to_real_path(params["lpNewFileName"])
    bFailIfExists = params["bFailIfExists"]

    if bFailIfExists and os.path.exists(lpNewFileName):
        return 0

    copyfile(lpExistingFileName, lpNewFileName)
    return 1

# BOOL SetFileAttributesA(
#   LPCSTR lpFileName,
#   DWORD  dwFileAttributes
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={
    "lpFileName": STRING,
    "dwFileAttributes": DWORD
})
def hook_SetFileAttributesA(ql, address, params):
    return 1



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#
# A Simple Windows Clipboard Simulation

NOT_LOCKED = -1


class Clipboard:

    def __init__(self, os):
        #super(Clipboard, self).__init__(ql)

        self.locked_by = NOT_LOCKED
        self.data = b"Default Clipboard Data"
        self.os = os
        # Valid formats taken from https://doxygen.reactos.org/d8/dd6/base_
        # 2applications_2mstsc_2constants_8h_source.html
        self.formats = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 128, 129, 130, 131, 142, 512, 767,
                        768, 1023]

    def open(self, h_wnd):
        """
        Lock clipboard to hWnd if not already locked. 
        If hWnd is null default to current thead id
        """
        if h_wnd == 0:
            hWnd = self.os.thread_manager.cur_thread.id

        if self.locked_by != NOT_LOCKED and self.locked_by != h_wnd:
            return 0
        else:
            self.locked_by = h_wnd
            return 1

    def format_available(self, fmt):
        if fmt in self.formats:
            return 1
        else:
            return 0

    def close(self):
        if self.locked_by == NOT_LOCKED:
            self.os.last_error = 0x58A  # ERROR_CLIPBOARD_NOT_OPEN
            return 0
        else:
            self.locked_by = NOT_LOCKED
            return 1

    def set_data(self, fmt, data):
        hWnd = self.os.thread_manager.cur_thread.id
        
        if self.locked_by != hWnd:
            self.os.last_error = 0x58A  # ERROR_CLIPBOARD_NOT_OPEN
            return 0
        else:
            if fmt not in self.formats:
                return 0
            self.data = data
            return 1

    def get_data(self, fmt):
        if fmt not in self.formats:
            return 0
        hWnd = self.os.thread_manager.cur_thread.id
        
        if self.locked_by != hWnd:
            self.os.last_error = 0x58A  # ERROR_CLIPBOARD_NOT_OPEN
            return 0
        else:
            return self.data



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall_time(ql, *args, **kw):
    regreturn = int(time.time())
    return regreturn


def ql_syscall_nanosleep(ql, nanosleep_req, nanosleep_rem, *args, **kw):
    def _sched_sleep(cur_thread):
        gevent.sleep(tv_sec)

    n = ql.archbit // 8 # 4 for 32-bit , 8 for 64-bit

    tv_sec = ql.unpack(ql.mem.read(nanosleep_req, n))
    tv_sec += ql.unpack(ql.mem.read(nanosleep_req + n, n)) / 1000000000

    if ql.os.thread_management == None:
        time.sleep(tv_sec)
    else:
        ql.emu_stop()
        ql.os.thread_management.cur_thread.sched_cb = _sched_sleep
        th = ql.os.thread_management.cur_thread

    regreturn = 0
    return regreturn


def ql_syscall_setitimer(ql, setitimer_which, setitimer_new_value, setitimer_old_value, *args, **kw):
    # TODO:The system provides each process with three interval timers, each decrementing in a distinct time domain.
    # When any timer expires, a signal is sent to the process, and the timer (potentially) restarts.
    # But I haven’t figured out how to send a signal yet.
    regreturn = 0
    return regreturn


def ql_syscall_times(ql, times_tbuf, *args, **kw):
    tmp_times = os.times()
    if times_tbuf != 0:
        tmp_buf = b''
        tmp_buf += ql.pack32(int(tmp_times.user * 1000))
        tmp_buf += ql.pack32(int(tmp_times.system * 1000))
        tmp_buf += ql.pack32(int(tmp_times.children_user * 1000))
        tmp_buf += ql.pack32(int(tmp_times.children_system * 1000))
        ql.mem.write(times_tbuf, tmp_buf)
    regreturn = int(tmp_times.elapsed * 100)
    return regreturn


def ql_syscall_gettimeofday(ql, gettimeofday_tv, gettimeofday_tz, *args, **kw):
    tmp_time = time.time()
    tv_sec = int(tmp_time)
    tv_usec = int((tmp_time - tv_sec) * 1000000)

    if gettimeofday_tv != 0:
        ql.mem.write(gettimeofday_tv, ql.pack32(tv_sec) + ql.pack32(tv_usec))
    if gettimeofday_tz != 0:
        ql.mem.write(gettimeofday_tz, b'\x00' * 8)
    regreturn = 0
    return regreturn



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import os

from qiling.exception import *
from qiling.os.stat import *

try:
    import fcntl
except ImportError:
    pass
import socket

class ql_file:
    def __init__(self, path, fd):
        self.__path = path
        self.__fd = fd
        # information for syscall mmap
        self._is_map_shared = False
        self._mapped_offset = -1

    @classmethod
    def open(self, open_path, open_flags, open_mode):
        open_mode &= 0x7fffffff

        try:
            fd = os.open(open_path, open_flags, open_mode)
        except OSError as e:
            raise QlSyscallError(e.errno, e.args[1] + ' : ' + e.filename)
        return self(open_path, fd)

    def read(self, read_len):
        return os.read(self.__fd, read_len)
    
    def write(self, write_buf):
        return os.write(self.__fd, write_buf)
    
    def fileno(self):
        return self.__fd
    
    def lseek(self, lseek_offset, lseek_origin = os.SEEK_SET):
        return os.lseek(self.__fd, lseek_offset, lseek_origin)
    
    def close(self):
        return os.close(self.__fd)
    
    def fstat(self):
        return Fstat(self.__fd)
    
    def ioctl(self, ioctl_cmd, ioctl_arg):
        try:
            return fcntl.ioctl(self.__fd, ioctl_cmd, ioctl_arg)
        except Exception:
            pass

    def tell(self):
        return self.lseek(0, os.SEEK_CUR)
    
    def dup(self):
        new_fd = os.dup(self.__fd)
        new_ql_file = ql_file(self.__path, new_fd)
        return new_ql_file
    
    def readline(self, end = b'\n'):
        ret = b''
        while True:
            c = self.read(1)
            ret += c
            if c == end:
                break
        return ret
    
    @property
    def name(self):
        return self.__path



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

import inspect


from .filestruct import ql_file
from .utils import QlOsUtils


# All mapped objects should inherit this class.
# Note this object is compatible with ql_file.
# Q: Why not derive from ql_file directly?
# A: ql_file assumes that it holds a path with a corresponding fd, but
#    a QlFsMappedObject doesn't have to be associated with a path or fd
#    and thus the default implementation may cause unexpected behaviors.
#    Simply let it crash if the method is not implemented.
#
#    A quick way to create a QlFsMappedObject is `ql_file.open` or `open`.
class QlFsMappedObject:
    def __init__(self):
        pass
    
    def read(self, expected_len):
        raise NotImplementedError("QlFsMappedObject method not implemented: read")
    
    def write(self, buffer):
        raise NotImplementedError("QlFsMappedObject method not implemented: write")
    
    def fileno(self):
        raise NotImplementedError("QlFsMappedObject method not implemented: fileno")
    
    def lseek(self, lseek_offset, lseek_origin):
        raise NotImplementedError("QlFsMappedObject method not implemented: lseek")
    
    def close(self):
        raise NotImplementedError("QlFsMappedObject method not implemented: close")
    
    def fstat(self):
        raise NotImplementedError("QlFsMappedObject method not implemented: fstat")
    
    def ioctl(self, ioctl_cmd, ioctl_arg):
        raise NotImplementedError("QlFsMappedObject method not implemented: ioctl")

    def tell(self):
        raise NotImplementedError("QlFsMappedObject method not implemented: tell")
    
    def dup(self):
        raise NotImplementedError("QlFsMappedObject method not implemented: dup")
    
    def readline(self, end = b'\n'):
        raise NotImplementedError("QlFsMappedObject method not implemented: readline")

    @property
    def name(self):
        raise NotImplementedError("QlFsMappedObject property not implemented: name")

class QlFsMapper:
    
    def __init__(self, ql):
        self._mapping = {}
        self._ql = ql
    
    @property
    def ql(self):
        return self._ql

    def _open_mapping_ql_file(self, ql_path, openflags, openmode):
        real_dest = self._mapping[ql_path]
        if isinstance(real_dest, str):
            qlfile = ql_file.open(real_dest, openflags, openmode)
            return qlfile
        elif inspect.isclass(real_dest):
            new_instance = real_dest()
            return new_instance
        else:
            return real_dest
    
    def _open_mapping(self, ql_path, openmode):
        real_dest = self._mapping[ql_path]
        if isinstance(real_dest, str):
            f = open(real_dest, openmode)
            return f
        elif inspect.isclass(real_dest):
            new_instance = real_dest()
            return new_instance
        else:
            return real_dest

    def has_mapping(self, fm):
        return fm in self._mapping

    def mapping_count(self):
        return len(self._mapping)

    def open_ql_file(self, path, openflags, openmode):
        if self.has_mapping(path):
            self.ql.log.info(f"mapping {path}")
            return self._open_mapping_ql_file(path, openflags, openmode)
        else:
            real_path = self.ql.os.transform_to_real_path(path)
            return ql_file.open(real_path, openflags, openmode)

    def open(self, path, openmode):
        if self.has_mapping(path):
            self.ql.log.info(f"mapping {path}")
            return self._open_mapping(path, openmode)
        else:
            real_path = self.ql.os.transform_to_real_path(path)
            return open(real_path, openmode)

    def _parse_path(self, p):
        if "__fspath__" in dir(p): # p is a os.PathLike object.
            p = p.__fspath__()
            if isinstance(p, bytes): # os.PathLike.__fspath__ may return bytes.
                p = p.decode("utf-8")
        return p

    # ql_path:   Emulated path which should be convertable to a string or a hashable object. e.g. pathlib.Path
    # real_dest: Mapped object, can be a string, an object or a class.
    #            string: mapped path in the host machine, e.g. `/dev/urandom` -> `/dev/urandom`.
    #            object: mapped object, will be returned each time the emulated path has been opened.
    #            class:  mapped class, will be used to create a new instance each time the emulated path has been opened.
    def add_fs_mapping(self, ql_path, real_dest):
        ql_path = self._parse_path(ql_path)
        real_dest = self._parse_path(real_dest)
        self._mapping[ql_path] = real_dest
        


#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from elftools.common.exceptions import ELFError
from elftools.common.py3compat import (
        ifilter, byte2int, bytes2str, itervalues, str2bytes, iterbytes)
from elftools.elf.elffile import ELFFile
from elftools.elf.dynamic import DynamicSection, DynamicSegment
from elftools.elf.enums import ENUM_D_TAG
from elftools.elf.segments import InterpSegment
from elftools.elf.sections import NoteSection, SymbolTableSection
from elftools.elf.gnuversions import (
    GNUVerSymSection, GNUVerDefSection,
    GNUVerNeedSection,
    )
from elftools.elf.relocation import RelocationSection
from elftools.elf.descriptions import (
    describe_ei_class, describe_ei_data, describe_ei_version,
    describe_ei_osabi, describe_e_type, describe_e_machine,
    describe_e_version_numeric, describe_p_type, describe_p_flags,
    describe_sh_type, describe_sh_flags,
    describe_symbol_type, describe_symbol_bind, describe_symbol_visibility,
    describe_symbol_shndx, describe_reloc_type, describe_dyn_tag,
    describe_dt_flags, describe_dt_flags_1, describe_ver_flags, describe_note,
    describe_attr_tag_arm
    )
from elftools.elf.constants import E_FLAGS
from elftools.elf.constants import E_FLAGS_MASKS

from qiling import Qiling


class QlReadELF(object):
    def __init__(self, ql:Qiling, elf_stream):
        self.ql = ql
        self.elffile = ELFFile(elf_stream)
        self._versioninfo = None

    def elf_file_header(self):
        elf_header = {}
        def add_info(key, value):
            elf_header[key] = value

        header = self.elffile.header
        e_ident = header['e_ident']

        add_info('Magic', ' '.join('%2.2x' % byte2int(b)
                   for b in self.elffile.e_ident_raw))
        add_info('Class',describe_ei_class(e_ident['EI_CLASS']))
        add_info('Data', describe_ei_data(e_ident['EI_DATA']))
        add_info('Version', e_ident['EI_VERSION'])
        add_info('OS/ABI', describe_ei_osabi(e_ident['EI_OSABI']))
        add_info('ABI Version', e_ident['EI_ABIVERSION'])
        add_info('Type', describe_e_type(header['e_type']))
        add_info('Machine', describe_e_machine(header['e_machine']))
        add_info('Version_e', describe_e_version_numeric(header['e_version']))
        add_info('Entry point address', self._format_hex(header['e_entry']))
        add_info('Start of program headers', header['e_phoff'])
        add_info('Start of section headers', header['e_shoff'])
        add_info('Flags', [self._format_hex(header['e_flags']),
                self.decode_flags(header['e_flags'])])
        add_info('Size of this header', header['e_ehsize'])
        add_info('Size of program headers', header['e_phentsize'])
        add_info('Number of program headers', header['e_phnum'])
        add_info('Size of section headers', header['e_shentsize'])
        add_info('Number of section headers', header['e_shnum'])
        add_info('Section header string table index', header['e_shstrndx'])

        return elf_header

    def elf_program_headers(self):
        program_headers = []
        def add_info(dic):
            program_headers.append(dic)

        if self.elffile.num_segments() == 0:
            return None

        for segment in self.elffile.iter_segments():
            program_hdr = {}
            program_hdr['Type'] = describe_p_type(segment['p_type'])
            program_hdr['Offset'] = self._format_hex(segment['p_offset'], fieldsize=6)
            program_hdr['VirtAddr'] = self._format_hex(segment['p_vaddr'], fullhex=True)
            program_hdr['PhysAddr'] = self._format_hex(segment['p_paddr'], fullhex=True)
            program_hdr['FileSiz'] = self._format_hex(segment['p_filesz'], fieldsize=5)
            program_hdr['MemSiz'] = self._format_hex(segment['p_memsz'], fieldsize=5)
            program_hdr['Flg'] = describe_p_flags(segment['p_flags'])
            program_hdr['Align'] = self._format_hex(segment['p_align'])

            add_info(program_hdr)

        return program_headers

    def elf_section_headers(self):
        section_headers = []
        def add_info(dic):
            section_headers.append(dic)

        if self.elffile.num_sections() == 0:
            return None

        for nsec, section in enumerate(self.elffile.iter_sections()):
            section_hdr = {}
            section_hdr['index'] = nsec
            section_hdr['Name'] = section.name
            section_hdr['Type'] = describe_sh_type(section['sh_type'])
            section_hdr['Addr'] = self._format_hex(section['sh_addr'], fieldsize=8, lead0x=False)
            section_hdr['Offset'] = self._format_hex(section['sh_offset'], fieldsize=6, lead0x=False)
            section_hdr['Size'] = self._format_hex(section['sh_size'], fieldsize=6, lead0x=False)
            section_hdr['ES'] = self._format_hex(section['sh_entsize'], fieldsize=2, lead0x=False)
            section_hdr['Flag'] = describe_sh_flags(section['sh_flags'])
            section_hdr['Lk'] = section['sh_link']
            section_hdr['Inf'] = section['sh_info']
            section_hdr['Al'] = section['sh_addralign']

            add_info(section_hdr)

        return section_headers

    def elf_symbol_tables(self):
        symbol_tables_list = []
        def add_info(dic):
            symbol_tables_list.append(dic)

        self._init_versioninfo()

        symbol_tables = [s for s in self.elffile.iter_sections()
                    if isinstance(s, SymbolTableSection)]

        if not symbol_tables and self.elffile.num_sections() == 0:
            return None

        for section in symbol_tables:
            if not isinstance(section, SymbolTableSection):
                continue

            if section['sh_entsize'] == 0:
                continue

            for nsym, symbol in enumerate(section.iter_symbols()):
                version_info = ''
                if (section['sh_type'] == 'SHT_DYNSYM' and
                        self._versioninfo['type'] == 'GNU'):
                    version = self._symbol_version(nsym)
                    if (version['name'] != symbol.name and
                        version['index'] not in ('VER_NDX_LOCAL',
                                                 'VER_NDX_GLOBAL')):
                        if version['filename']:
                            # external symbol
                            version_info = '@%(name)s (%(index)i)' % version
                        else:
                            # internal symbol
                            if version['hidden']:
                                version_info = '@%(name)s' % version
                            else:
                                version_info = '@@%(name)s' % version

                symbol_info = {}
                symbol_info['index'] = nsym
                symbol_info['Value'] = self._format_hex(
                        symbol['st_value'], fullhex=True, lead0x=False)
                symbol_info['Size'] = symbol['st_size']
                symbol_info['Type'] = describe_symbol_type(symbol['st_info']['type'])
                symbol_info['Bind'] = describe_symbol_bind(symbol['st_info']['bind'])
                symbol_info['Vis'] = describe_symbol_visibility(symbol['st_other']['visibility'])
                symbol_info['Ndx'] = describe_symbol_shndx(symbol['st_shndx'])
                symbol_info['Name'] = symbol.name
                symbol_info['version_info'] = version_info
                add_info(symbol_info)
        return symbol_tables_list

    def decode_flags(self, flags):
        description = ""
        if self.elffile['e_machine'] == "EM_ARM":
            eabi = flags & E_FLAGS.EF_ARM_EABIMASK
            flags &= ~E_FLAGS.EF_ARM_EABIMASK

            if flags & E_FLAGS.EF_ARM_RELEXEC:
                description += ', relocatable executabl'
                flags &= ~E_FLAGS.EF_ARM_RELEXEC

            if eabi == E_FLAGS.EF_ARM_EABI_VER5:
                EF_ARM_KNOWN_FLAGS = E_FLAGS.EF_ARM_ABI_FLOAT_SOFT|E_FLAGS.EF_ARM_ABI_FLOAT_HARD|E_FLAGS.EF_ARM_LE8|E_FLAGS.EF_ARM_BE8
                description += ', Version5 EABI'
                if flags & E_FLAGS.EF_ARM_ABI_FLOAT_SOFT:
                    description += ", soft-float ABI"
                elif flags & E_FLAGS.EF_ARM_ABI_FLOAT_HARD:
                    description += ", hard-float ABI"

                if flags & E_FLAGS.EF_ARM_BE8:
                    description += ", BE8"
                elif flags & E_FLAGS.EF_ARM_LE8:
                    description += ", LE8"

                if flags & ~EF_ARM_KNOWN_FLAGS:
                    description += ', <unknown>'
            else:
                description += ', <unrecognized EABI>'

        elif self.elffile['e_machine'] == "EM_MIPS":
            if flags & E_FLAGS.EF_MIPS_NOREORDER:
                description += ", noreorder"
            if flags & E_FLAGS.EF_MIPS_PIC:
                description += ", pic"
            if flags & E_FLAGS.EF_MIPS_CPIC:
                description += ", cpic"
            if (flags & E_FLAGS.EF_MIPS_ABI2):
                description += ", abi2"
            if (flags & E_FLAGS.EF_MIPS_32BITMODE):
                description += ", 32bitmode"
            if (flags & E_FLAGS_MASKS.EFM_MIPS_ABI_O32):
                description += ", o32"
            elif (flags & E_FLAGS_MASKS.EFM_MIPS_ABI_O64):
                description += ", o64"
            elif (flags & E_FLAGS_MASKS.EFM_MIPS_ABI_EABI32):
                description += ", eabi32"
            elif (flags & E_FLAGS_MASKS.EFM_MIPS_ABI_EABI64):
                description += ", eabi64"
            if (flags & E_FLAGS.EF_MIPS_ARCH) == E_FLAGS.EF_MIPS_ARCH_1:
                description += ", mips1"
            if (flags & E_FLAGS.EF_MIPS_ARCH) == E_FLAGS.EF_MIPS_ARCH_2:
                description += ", mips2"
            if (flags & E_FLAGS.EF_MIPS_ARCH) == E_FLAGS.EF_MIPS_ARCH_3:
                description += ", mips3"
            if (flags & E_FLAGS.EF_MIPS_ARCH) == E_FLAGS.EF_MIPS_ARCH_4:
                description += ", mips4"
            if (flags & E_FLAGS.EF_MIPS_ARCH) == E_FLAGS.EF_MIPS_ARCH_5:
                description += ", mips5"
            if (flags & E_FLAGS.EF_MIPS_ARCH) == E_FLAGS.EF_MIPS_ARCH_32R2:
                description += ", mips32r2"
            if (flags & E_FLAGS.EF_MIPS_ARCH) == E_FLAGS.EF_MIPS_ARCH_64R2:
                description += ", mips64r2"
            if (flags & E_FLAGS.EF_MIPS_ARCH) == E_FLAGS.EF_MIPS_ARCH_32:
                description += ", mips32"
            if (flags & E_FLAGS.EF_MIPS_ARCH) == E_FLAGS.EF_MIPS_ARCH_64:
                description += ", mips64"

        return description

    def _format_hex(self, addr, fieldsize=None, fullhex=False, lead0x=True,
                    alternate=False):
        """ Format an address into a hexadecimal string.

            fieldsize:
                Size of the hexadecimal field (with leading zeros to fit the
                address into. For example with fieldsize=8, the format will
                be %08x
                If None, the minimal required field size will be used.

            fullhex:
                If True, override fieldsize to set it to the maximal size
                needed for the elfclass

            lead0x:
                If True, leading 0x is added

            alternate:
                If True, override lead0x to emulate the alternate
                hexadecimal form specified in format string with the #
                character: only non-zero values are prefixed with 0x.
                This form is used by readelf.
        """
        if alternate:
            if addr == 0:
                lead0x = False
            else:
                lead0x = True
                fieldsize -= 2

        s = '0x' if lead0x else ''
        if fullhex:
            fieldsize = 8 if self.elffile.elfclass == 32 else 16
        if fieldsize is None:
            field = '%x'
        else:
            field = '%' + '0%sx' % fieldsize
        return s + field % addr

    def _init_versioninfo(self):
        """ Search and initialize informations about version related sections
            and the kind of versioning used (GNU or Solaris).
        """
        if self._versioninfo is not None:
            return

        self._versioninfo = {'versym': None, 'verdef': None,
                             'verneed': None, 'type': None}

        for section in self.elffile.iter_sections():
            if isinstance(section, GNUVerSymSection):
                self._versioninfo['versym'] = section
            elif isinstance(section, GNUVerDefSection):
                self._versioninfo['verdef'] = section
            elif isinstance(section, GNUVerNeedSection):
                self._versioninfo['verneed'] = section
            elif isinstance(section, DynamicSection):
                for tag in section.iter_tags():
                    if tag['d_tag'] == 'DT_VERSYM':
                        self._versioninfo['type'] = 'GNU'
                        break

        if not self._versioninfo['type'] and (
                self._versioninfo['verneed'] or self._versioninfo['verdef']):
            self._versioninfo['type'] = 'Solaris'

    def _symbol_version(self, nsym):
        """ Return a dict containing information on the
                   or None if no version information is available
        """
        self._init_versioninfo()

        symbol_version = dict.fromkeys(('index', 'name', 'filename', 'hidden'))

        if (not self._versioninfo['versym'] or
                nsym >= self._versioninfo['versym'].num_symbols()):
            return None

        symbol = self._versioninfo['versym'].get_symbol(nsym)
        index = symbol.entry['ndx']
        if not index in ('VER_NDX_LOCAL', 'VER_NDX_GLOBAL'):
            index = int(index)

            if self._versioninfo['type'] == 'GNU':
                # In GNU versioning mode, the highest bit is used to
                # store wether the symbol is hidden or not
                if index & 0x8000:
                    index &= ~0x8000
                    symbol_version['hidden'] = True

            if (self._versioninfo['verdef'] and
                    index <= self._versioninfo['verdef'].num_versions()):
                _, verdaux_iter = \
                        self._versioninfo['verdef'].get_version(index)
                symbol_version['name'] = next(verdaux_iter).name
            else:
                verneed, vernaux = \
                        self._versioninfo['verneed'].get_version(index)
                symbol_version['name'] = vernaux.name
                symbol_version['filename'] = verneed.name

        symbol_version['index'] = index
        return symbol_version



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.os.windows.fncc import *
from qiling.exception import *

dllname = 'kernel32_dll'

# HANDLE WINAPI GetStdHandle(
#   _In_ DWORD nStdHandle
# );
@winsdkapi(cc=STDCALL, dllname=dllname, replace_params={"nStdHandle": DWORD})
def hook_GetStdHandle(ql, address, params):
    nStdHandle = params["nStdHandle"]
    return nStdHandle


# LPSTR GetCommandLineA(
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetCommandLineA(ql, address, params):
    cmdline = ql.loader.cmdline + b"\x00"
    addr = ql.os.heap.alloc(len(cmdline))
    ql.mem.write(addr, cmdline)
    return addr


# LPSTR GetCommandLineW(
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetCommandLineW(ql, address, params):
    cmdline = ql.loader.cmdline.decode('ascii').encode('utf-16le')
    addr = ql.os.heap.alloc(len(cmdline))
    ql.mem.write(addr, cmdline)
    return addr


# LPWCH GetEnvironmentStrings(
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetEnvironmentStrings(ql, address, params):
    cmdline = b"\x00"
    addr = ql.os.heap.alloc(len(cmdline))
    ql.mem.write(addr, cmdline)
    return addr


# LPWCH GetEnvironmentStringsW(
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetEnvironmentStringsW(ql, address, params):
    cmdline = b"\x00\x00"
    addr = ql.os.heap.alloc(len(cmdline))
    ql.mem.write(addr, cmdline)
    return addr


# BOOL FreeEnvironmentStringsW(
#   LPWCH penv
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_FreeEnvironmentStringsW(ql, address, params):
    ret = 1
    return ret


# DWORD ExpandEnvironmentStringsW(
#   LPCWSTR lpSrc,
#   LPWSTR  lpDst,
#   DWORD   nSize
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_ExpandEnvironmentStringsW(ql, address, params):
    string: str = params["lpSrc"]
    start = string.find("%")
    end = string.rfind("%")
    substring = string[start + 1:end]
    result = ql.os.profile["PATH"].get(substring, None)
    if result is None:
        ql.log.debug(substring)
        raise QlErrorNotImplemented("API not implemented")
    result = (string[:start] + result + string[end + 1:] + "\x00").encode("utf-16le")
    dst = params["lpDst"]
    max_size = params["nSize"]
    if len(result) <= max_size:
        ql.mem.write(dst, result)
    return len(result)


# DWORD GetEnvironmentVariableA(
#   LPCSTR lpName,
#   LPSTR  lpBuffer,
#   DWORD  nSize
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetEnvironmentVariableA(ql, address, params):
    ret = 0
    return ret

# DWORD GetEnvironmentVariableW(
#   LPCWSTR lpName,
#   LPWSTR  lpBuffer,
#   DWORD  nSize
# );
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetEnvironmentVariableW(ql, address, params):
    ret = 0
    return ret



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

try:
    import resource
except ImportError:
    # The library 'resource' does not exist on windows, so provide a dummy shim
    class DummyResource:
        def getrlimit(self, resource):
            return (-1, -1)
        def setrlimit(self, resource, rlim):
            pass
    resource = DummyResource()


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall_ugetrlimit(ql, ugetrlimit_resource, ugetrlimit_rlim, *args, **kw):
    rlim = resource.getrlimit(ugetrlimit_resource)
    ql.mem.write(ugetrlimit_rlim, ql.pack32s(rlim[0]) + ql.pack32s(rlim[1]))
    regreturn = 0
    return regreturn


def ql_syscall_setrlimit(ql, setrlimit_resource, setrlimit_rlim, *args, **kw):
    # maybe we can nop the setrlimit
    tmp_rlim = (ql.unpack32s(ql.mem.read(setrlimit_rlim, 4)), ql.unpack32s(ql.mem.read(setrlimit_rlim + 4, 4)))
    resource.setrlimit(setrlimit_resource, tmp_rlim)
    regreturn = 0
    return regreturn


def ql_syscall_prlimit64(ql, prlimit64_pid, prlimit64_resource, prlimit64_new_limit, prlimit64_old_limit, *args, **kw):
    # setrlimit() and getrlimit()
    if prlimit64_pid == 0 and prlimit64_new_limit == 0:
        rlim = resource.getrlimit(prlimit64_resource)
        ql.mem.write(prlimit64_old_limit, ql.packs(rlim[0]) + ql.packs(rlim[1]))
        regreturn = 0
    else:
        # set other process which pid != 0
        regreturn = -1
    return regreturn


def ql_syscall_getpriority(ql, getpriority_which, getpriority_who, null1, null2, null3, null4):
    base = os.getpriority(getpriority_which, getpriority_who)
    return base



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from .const import *
from qiling.const import *


def _invert_dict(d):
    return { v:k for k, v in d.items()}


def _constant_mapping(bits, d_map, ret=None, single_mapping=False):
    if ret is None:
        ret = []

    b_map = _invert_dict(d_map)

    if single_mapping:
        return b_map[bits]

    for val, sym in b_map.items():
        if val & bits != 0:
            bits ^= val
            ret.append(sym)

    if bits != 0:
        ret.append(str(bits))

    return " | ".join(ret)

#
#

def ql_open_flag_mapping(ql, flags):
    def flag_mapping(flags, mapping_name, mapping_from, mapping_to):
        ret = 0
        for n in mapping_name:
            if  (flags & mapping_from[n]) == mapping_from[n]:
                ret = ret | mapping_to[n]
        return ret

    open_flags_name = [
        "O_RDONLY",
        "O_WRONLY",
        "O_RDWR",
        "O_NONBLOCK",
        "O_APPEND",
        "O_ASYNC",
        "O_SYNC",
        "O_NOFOLLOW",
        "O_CREAT",
        "O_TRUNC",
        "O_EXCL",
        "O_NOCTTY",
        "O_DIRECTORY",
        "O_BINARY"
    ]

    mac_open_flags = {
        "O_RDONLY": 0x0000,
        "O_WRONLY": 0x0001,
        "O_RDWR": 0x0002,
        "O_NONBLOCK": 0x0004,
        "O_APPEND": 0x0008,
        "O_ASYNC": 0x0040,
        "O_SYNC": 0x0080,
        "O_NOFOLLOW": 0x0100,
        "O_CREAT": 0x0200,
        "O_TRUNC": 0x0400,
        "O_EXCL": 0x0800,
        "O_NOCTTY": 0x20000,
        "O_DIRECTORY": 0x100000,
        "O_BINARY": 0
    }

    linux_open_flags = {
        'O_RDONLY': 0,
        'O_WRONLY': 1,
        'O_RDWR': 2,
        'O_NONBLOCK': 2048,
        'O_APPEND': 1024,
        'O_ASYNC': 8192,
        'O_SYNC': 1052672,
        'O_NOFOLLOW': 131072,
        'O_CREAT': 64,
        'O_TRUNC': 512,
        'O_EXCL': 128,
        'O_NOCTTY': 256,
        'O_DIRECTORY': 65536,
        'O_BINARY': 0
    }

    mips_open_flags = {
        'O_RDONLY': 0x0,
        'O_WRONLY': 0x1,
        'O_RDWR': 0x2,
        'O_NONBLOCK': 0x80,
        'O_APPEND': 0x8,
        'O_ASYNC': 0x1000,
        'O_SYNC': 0x4000,
        'O_NOFOLLOW': 0x20000,
        'O_CREAT': 0x100,
        'O_TRUNC': 0x200,
        'O_EXCL': 0x400,
        'O_NOCTTY': 0x800,
        'O_DIRECTORY': 0x100000,
        'O_BINARY' : 0
    }

    windows_open_flags = {
        'O_RDONLY': 0,
        'O_WRONLY': 1,
        'O_RDWR': 2,
        'O_NONBLOCK': 2, # Windows doesn't have a corresponding one, assume RW
        'O_APPEND': 8,
        'O_ASYNC': 2, # Windows doesn't have a corresponding one, assume RW
        'O_SYNC': 2, # Windows doesn't have a corresponding one, assume RW
        'O_NOFOLLOW': 2, # Windows doesn't have a corresponding one, assume RW
        'O_CREAT': 256,
        'O_TRUNC': 512,
        'O_EXCL': 1024,
        'O_NOCTTY': 2,
        'O_DIRECTORY': 2, # Windows doesn't have a corresponding one, assume RW
        'O_BINARY': 32768
    }

    f = {}
    t = {}

    if ql.archtype != QL_ARCH.MIPS:
        if ql.platform == None or ql.platform == ql.ostype:
            return flags
    
        if ql.ostype == QL_OS.LINUX:
            f = linux_open_flags
        elif ql.ostype == QL_OS.MACOS:
            f = mac_open_flags
        elif ql.ostype == QL_OS.WINDOWS:
            f = windows_open_flags
        
        if ql.platform == QL_OS.WINDOWS:
            t = windows_open_flags
        elif ql.platform == QL_OS.MACOS:
            t = mac_open_flags
        elif ql.platform == QL_OS.LINUX:
            t = linux_open_flags
    
    elif ql.archtype == QL_ARCH.MIPS and ql.platform == QL_OS.LINUX:
        f = mips_open_flags
        t = linux_open_flags
    elif ql.archtype == QL_ARCH.MIPS and ql.platform == QL_OS.MACOS:
        f = mips_open_flags
        t = mac_open_flags
    elif ql.archtype == QL_ARCH.MIPS and ql.platform == QL_OS.WINDOWS:
        f = mips_open_flags
        t = windows_open_flags        

    return flag_mapping(flags, open_flags_name, f, t)


def open_flags_mapping(flags, arch):

    arm64_open_flags = arm_open_flags

    mapping_dict = {
            QL_ARCH.X86: linux_open_flags,
            QL_ARCH.X8664: linux_open_flags,
            QL_ARCH.ARM: arm_open_flags,
            QL_ARCH.ARM64: arm64_open_flags,
            QL_ARCH.MIPS: mips_open_flags,
            QL_OS.MACOS: mac_open_flags,
            }.get(arch)

    ret = ["O_RDONLY"]

    return _constant_mapping(flags, mapping_dict, ret)


def mmap_flag_mapping(flags):

    mmap_flags = {
        'MAP_SHARED'    : 0x01,
        'MAP_PRIVATE'   : 0x02,
        'MAP_FIXED'     : 0x10,
        'MAP_ANONYMOUS' : 0x20,
    }

    return _constant_mapping(flags, mmap_flags)


def mmap_prot_mapping(prots):

    if prots == 0x0:
        return 'PROT_NONE'

    mmap_prots = {
        'PROT_READ' : 0x1,
        'PROT_WRITE': 0x2,
        'PROT_EXEC' : 0x4,
    }

    return _constant_mapping(prots, mmap_prots)


def socket_type_mapping(t, arch):
    socket_type_map = {
            QL_ARCH.X86: linux_socket_types,
            QL_ARCH.X8664: linux_socket_types,
            QL_ARCH.ARM: arm_socket_types,
            QL_ARCH.ARM_THUMB: arm_socket_types,
            QL_ARCH.ARM64: arm_socket_types,
            QL_ARCH.MIPS: mips_socket_types,
            QL_OS.MACOS: linux_socket_types,
            }.get(arch)

    return _constant_mapping(t, socket_type_map)


def socket_domain_mapping(p, arch):
    socket_domain_map = {
            QL_ARCH.X86: linux_socket_domain,
            QL_ARCH.X8664: linux_socket_domain,
            QL_ARCH.ARM: arm_socket_domain,
            QL_ARCH.ARM_THUMB: arm_socket_domain,
            QL_ARCH.ARM64: arm_socket_domain,
            QL_ARCH.MIPS: mips_socket_domain,
            QL_OS.MACOS: "macos_socket_domain",
            }.get(arch)

    return _constant_mapping(p, socket_domain_map, single_mapping=True)



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from qiling.os.uefi.const import EFI_SUCCESS, EFI_NOT_FOUND, EFI_UNSUPPORTED, EFI_BUFFER_TOO_SMALL, EFI_INVALID_PARAMETER
from qiling.os.uefi.utils import read_int64, write_int64, check_and_notify_protocols, signal_event
from qiling.os.uefi.UefiSpec import EFI_LOCATE_SEARCH_TYPE

# TODO: get rid of this
pointer_size = 8

def LocateHandles(context, params):
	SearchType = params["SearchType"]
	Protocol = params["Protocol"]

	# get all handles
	if SearchType == EFI_LOCATE_SEARCH_TYPE.AllHandles:
		handles = context.protocols.keys()

	# get all handles that support the specified protocol
	elif SearchType == EFI_LOCATE_SEARCH_TYPE.ByProtocol:
		handles = [handle for handle, guid_dic in context.protocols.items() if Protocol in guid_dic]

	else:
		handles = []

	return len(handles) * pointer_size, handles

def InstallProtocolInterface(context, params):
	handle = read_int64(context.ql, params["Handle"])

	if handle == 0:
		handle = context.heap.alloc(1)

	dic = context.protocols.get(handle, {})

	dic[params["Protocol"]] = params["Interface"]
	context.protocols[handle] = dic

	for (event_id, event_dic) in context.ql.loader.events.items():
		if event_dic['Guid'] == params['Protocol']:
			# The event was previously registered by 'RegisterProtocolNotify'.
			signal_event(context.ql, event_id)	

	check_and_notify_protocols(context.ql)
	write_int64(context.ql, params["Handle"], handle)

	return EFI_SUCCESS

def ReinstallProtocolInterface(context, params):
	handle = params["Handle"]

	if handle not in context.protocols:
		return EFI_NOT_FOUND

	dic = context.protocols[handle]
	protocol = params["Protocol"]

	if protocol not in dic:
		return EFI_NOT_FOUND

	dic[protocol] = params["NewInterface"]

	return EFI_SUCCESS

def UninstallProtocolInterface(context, params):
	handle = params["Handle"]

	if handle not in context.protocols:
		return EFI_NOT_FOUND

	dic = context.protocols[handle]
	protocol = params["Protocol"]

	if protocol not in dic:
		return EFI_NOT_FOUND

	del dic[protocol]

	return EFI_SUCCESS

def HandleProtocol(context, params):
	handle = params["Handle"]
	protocol = params["Protocol"]
	interface = params['Interface']

	if handle in context.protocols:
		supported = context.protocols[handle]

		if protocol in supported:
			write_int64(context.ql, interface, supported[protocol])

			return EFI_SUCCESS

	return EFI_UNSUPPORTED

def LocateHandle(context, params):
	buffer_size, handles = LocateHandles(context, params)

	if len(handles) == 0:
		return EFI_NOT_FOUND

	ret = EFI_BUFFER_TOO_SMALL

	if read_int64(context.ql, params["BufferSize"]) >= buffer_size:
		ptr = params["Buffer"]

		for handle in handles:
			write_int64(context.ql, ptr, handle)
			ptr += pointer_size

		ret = EFI_SUCCESS

	write_int64(context.ql, params["BufferSize"], buffer_size)

	return ret

def LocateProtocol(context, params):
	protocol = params['Protocol']

	for handle, guid_dic in context.protocols.items():
		if "Handle" in params and params["Handle"] != handle:
			continue

		if protocol in guid_dic:
			# write protocol address to out variable Interface
			write_int64(context.ql, params['Interface'], guid_dic[protocol])
			return EFI_SUCCESS

	context.ql.log.warning(f'protocol with guid {protocol} not found')

	return EFI_NOT_FOUND

def InstallConfigurationTable(context, params):
	guid = params["Guid"]
	table = params["Table"]

	if not guid:
		return EFI_INVALID_PARAMETER
	
	context.install_configuration_table(guid, table)

	return EFI_SUCCESS



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from configparser import ConfigParser
import ctypes, ntpath, os, pickle, platform
import io
from sys import stdin, stdout
# See https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports
from typing import Dict, List, Union
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .arch.register import QlRegisterManager
    from .arch.arch import QlArch
    from .os.os import QlOs
    from .os.memory import QlMemoryManager
    from .loader.loader import QlLoader

from .const import QL_ARCH_ENDIAN, QL_ENDIAN, QL_INTERCEPT, QL_OS_POSIX, QL_OS_ALL, QL_OUTPUT, QL_OS
from .exception import QlErrorFileNotFound, QlErrorArch, QlErrorOsType, QlErrorOutput
from .utils import *
from .core_struct import QlCoreStructs
from .core_hooks import QlCoreHooks
from .__version__ import __version__

# Mixin Pattern
class Qiling(QlCoreHooks, QlCoreStructs):    
    def __init__(
            self,
            argv=None,
            rootfs=None,
            env=None,
            code=None,
            shellcoder=None,
            ostype=None,
            archtype=None,
            bigendian=False,
            output=None,
            verbose=1,
            profile=None,
            console=True,
            log_file=None,
            log_override=None,
            libcache = False,
            multithread = False,
            filters = None,
            stop_on_stackpointer = False,
            stop_on_exit_trap = False,
            stdin=0,
            stdout=0,
            stderr=0,
    ):
        """ Create a Qiling instance.

            For each argument or property, please refer to its docstring. e.g. Qiling.multithread.__doc__

            The only exception is "bigendian" parameter, see Qiling.archendian.__doc__ for details.
        """

        ##################################
        # Definition during ql=Qiling()  #
        ##################################
        self._argv = argv
        self._rootfs = rootfs
        self._env = env if env else {}
        self._code = code
        self._shellcoder = shellcoder
        self._ostype = ostype
        self._archtype = archtype
        self._archendian = None
        self._archbit = None
        self._pointersize = None
        self._profile = profile
        self._console = console
        self._log_file = log_file
        self._multithread = multithread
        self._log_file_fd = None
        self._log_filter = None
        self._log_override = log_override
        self._filters = filters
        self._platform = ostype_convert(platform.system().lower())
        self._internal_exception = None
        self._uc = None
        self._stop_options = QlStopOptions(stackpointer=stop_on_stackpointer, exit_trap=stop_on_exit_trap)
        
        ##################################
        # Definition after ql=Qiling()   #
        ##################################
        self._stdin = stdin
        self._stdout = stdout
        self._stderr = stderr
        self._output = output
        self._verbose = verbose
        self._libcache = libcache
        self._patch_bin = []
        self._patch_lib = []
        self._debug_stop = False
        self._debugger = None
        self._root = False

        ###############################
        # Properties configured later #
        ###############################
        self.entry_point = None
        self.exit_point = None
        self.timeout = None
        self.count = None
        self._initial_sp = None


        """
        Qiling Framework Core Engine
        """
        ##############
        # Shellcode? #
        ##############

        # for Legacy
        if self._shellcoder:
            self._code = self._shellcoder

        if self._code or (self._archtype and type(self._archtype) == str):
            if (self._archtype and type(self._archtype) == str):
                self._archtype = arch_convert(self._archtype.lower())
            
            if (self._ostype and type(self._ostype) == str):     
                self._ostype = ostype_convert(self._ostype.lower())
            
            self._argv = ["qilingcode"]
            if self._rootfs is None:
                self._rootfs = "."                

        # file check
        if self._code is None:
            if not os.path.exists(str(self._argv[0])):
                raise QlErrorFileNotFound("Target binary not found")
            if not os.path.exists(self._rootfs):
                raise QlErrorFileNotFound("Target rootfs not found")
        
        self._path = (str(self._argv[0]))
        self._targetname = ntpath.basename(self._argv[0])

        ##########
        # Loader #
        ##########
        if self._code is None:
            guessed_archtype, guessed_ostype, guessed_archendian = ql_guess_emu_env(self._path)
            if self._ostype is None:
                self._ostype = guessed_ostype
            if self._archtype is None:
                self._archtype = guessed_archtype
            if self.archendian is None:
                self._archendian = guessed_archendian

            if not ql_is_valid_ostype(self._ostype):
                raise QlErrorOsType("Invalid OSType")

            if not ql_is_valid_arch(self._archtype):
                raise QlErrorArch("Invalid Arch %s" % self._archtype)

        self._loader = loader_setup(self._ostype, self)

        #####################
        # Profile & Logging #
        #####################
        self._profile, debugmsg = profile_setup(self.ostype, self.profile, self)

        # Log's configuration
        
        # Setup output mode.
        self._output = output_convert(self._output)

        self._log_file_fd, self._log_filter = ql_setup_logger(self,
                                                              self._log_file,
                                                              self._console, 
                                                              self._filters, 
                                                              self._multithread,
                                                              self._log_override)

        self.log.setLevel(ql_resolve_logger_level(self._output, self._verbose))

        # Now that the logger is configured, we can log profile debug msg:
        self.log.debug(debugmsg)

        ########################
        # Archbit & Endianness #
        ########################
        self._archbit = ql_get_arch_bits(self._archtype)
        self._pointersize = (self.archbit // 8)  
        
        # Endian for shellcode needs to set manually
        if self._code:
            self._archendian = QL_ENDIAN.EL
            if bigendian == True and self._archtype in (QL_ARCH_ENDIAN):
                self._archendian = QL_ENDIAN.EB
        
        # Once we finish setting up archendian and arcbit, we can init QlCoreStructs.
        QlCoreStructs.__init__(self, self._archendian, self._archbit)

        ##############
        # Components #
        ##############
        self._mem = component_setup("os", "memory", self)
        self._reg = component_setup("arch", "register", self)
        self._arch = arch_setup(self.archtype, self)
        
        # Once we finish setting up arch layer, we can init QlCoreHooks.
        self.uc = self.arch.init_uc
        QlCoreHooks.__init__(self, self.uc)

        self._os = os_setup(self.archtype, self.ostype, self)

        # Run the loader
        self.loader.run()
        
        # Setup Outpt
        self.os.setup_output()

        # Add extra guard options when configured to do so
        self._init_stop_guard()

    
    #####################
    # Qiling Components #
    #####################

    @property
    def mem(self) -> "QlMemoryManager":
        """ Qiling memory manager.

            Example: ql.mem.read(0xdeadbeaf, 4)
        """
        return self._mem
    
    @property
    def reg(self) -> "QlRegisterManager":
        """ Qiling register manager.

            Example: ql.reg.eax = 1
        """
        return self._reg
    
    @property
    def arch(self) -> "QlArch":
        """ Qiling architecture layer.

            Also see qiling/arch/<arch>.py
        """
        return self._arch

    @property
    def loader(self) -> "QlLoader":
        """ Qiling loader layer.

            Also see qiling/loader/<filetype>.py
        """
        return self._loader

    @property
    def os(self) -> "QlOs":
        """ Qiling os layer.

            Also see qiling/os/<os>/<os>/py
        """
        return self._os

    @property
    def log(self) -> logging.Logger:
        """ Returns the logger this Qiling instance uses.

            You can override this log by passing `log_override=your_log` to Qiling.__init__

            Type: logging.Logger
            Example: ql.log.info("This goes to terminal")
        """
        return self._log_file_fd

    ##################
    # Qiling Options #
    ##################

    # If an option doesn't have a setter, it means that it can be only set during Qiling.__init__
    @property
    def console(self) -> bool:
        """ Specify whether enabling console output. 

            Type: bool
            Example: Qiling(console=True)
        """
        return self._console

    @property
    def log_file(self) -> str:
        """ Log to a file.

            Type: str
            Example: Qiling(log_file="./ql.log")
        """
        return self._log_file

    @property
    def multithread(self) -> bool:
        """ Specify whether multithread has been enabled.

            WARNING: This property shouldn't be set after Qiling.__init__.

            Type: bool
            Example: Qiling(multithread=True)
        """
        return self._multithread

    @property
    def profile(self) -> ConfigParser:
        """ Program profile. See qiling/profiles/*.ql for details.

            Note: Please pass None or the path string to Qiling.__init__.

            Type: ConfigParser
            Value: str
            Example: Qiling(profile="profiles/dos.ql")
        """
        return self._profile
    
    @property
    def argv(self) -> List[str]:
        """ The program argv.

            Type: List[str]
            Example: Qiling(argv=['/bin/ls', '-a'])
        """
        return self._argv

    @property
    def rootfs(self) -> str:
        """ The program rootfs. For some common rootfs, see examples/rootfs/ for details.

            Type: str
            Example: Qiling(argv=['/bin/ls', '-a'], rootfs='examples/rootfs/x8664_linux/')
        """
        return self._rootfs

    @property
    def env(self) -> Dict[str, str]:
        """ The program environment variables.

            Type: Dict[str, str]
            Example: Qiling(env={"LC_ALL" : "en_US.UTF-8"})
        """
        return self._env

    @property
    def ostype(self) -> int:
        """ The emulated os type.

            Note: Please pass None or one of the strings below to Qiling.__init__.
                  If you use shellcode, you must specify ostype and archtype manually.

            Type: int.
            Values:
              - "macos" : macOS.
              - "darwin" : an alias to "macos".
              - "freebsd" : FreeBSD
              - "windows" : Windows
              - "uefi" : UEFI
              - "dos" : DOS
            Example: Qiling(code=b"\x90", ostype="macos", archtype="x8664", bigendian=False)
        """
        return self._ostype

    @property
    def archtype(self) -> int:
        """ The emulated architecture type.

            Note: Please pass None or one of the strings below to Qiling.__init__.
                  If you use shellcode, you must specify ostype and archtype manually.
            
            Type: int
            Values:
              - "x86" : x86_32
              - "x8664" : x86_64
              - "mips" : MIPS
              - "arm" : ARM
              - "arm_thumb" : ARM with thumb mode.
              - "arm64" : ARM64
              - "a8086" : 8086
            Example: Qiling(code=b"\x90", ostype="macos", archtype="x8664", bigendian=False)
        """
        return self._archtype

    @property
    def archendian(self) -> int:
        """ The architecure endian.

            Note: Please pass "bigendian=True" or "bingendian=False" to set this property.
                  This option only takes effect for shellcode.

            Type: int
            Example: Qiling(code=b"\x90", ostype="macos", archtype="x8664", bigendian=False)
        """
        return self._archendian

    @property
    def archbit(self) -> int:
        """ The bits of the current architecutre.

            Type: int
        """
        return self._archbit

    @property
    def pointersize(self) -> int:
        """ The pointer size of current architecture.

            Type: int
        """
        return self._pointersize

    @property
    def code(self) -> bytes:
        """ The shellcode to execute.

            Note: It can't be used with "argv" parameter.

            Type: bytes
            Example: Qiling(code=b"\x90", ostype="macos", archtype="x8664", bigendian=False)
        """
        return self._code

    @property
    def path(self) -> str:
        """ The file path of the executable.

            Type: str
        """
        return self._path
    
    @property
    def targetname(self) -> str:
        """ The target name of the executable. e.g. "c.exe" in "a\b\c.exe"

            Type: str
        """
        return self._targetname

    @property
    def platform(self):
        """ Specify current platform where Qiling runs on.

            Type: int
            Values: All possible values from platform.system()
        """
        return self._platform

    @platform.setter
    def platform(self, value):
        if type(value) is str:
            self._platform = ostype_convert(value.lower())
        else:
            self._platform = value

    @property
    def internal_exception(self) -> Exception:
        """ Internal exception catched during Unicorn callback. Not intended for regular users.

            Type: Exception
        """
        return self._internal_exception

    @property
    def stdin(self) -> io.IOBase:
        """ Stdin of the program. Can be any object which implements (even part of) io.IOBase.

            Type: io.Base
            Example: - ql = Qiling(stdin=sys.stdin)
                     - ql.stdin = sys.stdin
        """
        return self._stdin
    
    @stdin.setter
    def stdin(self, s):
        self._stdin = s
    
    @property
    def stdout(self) -> io.IOBase:
        """ Stdout of the program. Can be any object which implements (even part of) io.IOBase.

            Type: io.Base
            Example: - ql = Qiling(stdout=sys.stdout)
                     - ql.stdout = sys.stdout
        """
        return self._stdout

    @stdout.setter
    def stdout(self, s):
        self._stdout = s
    
    @property
    def stderr(self) -> io.IOBase:
        """ Stdout of the program. Can be any object which implements (even part of) io.IOBase.

            Type: io.Base
            Example: - ql = Qiling(stderr=sys.stderr)
                     - ql.stderr = sys.stderr
        """
        return self._stderr
    
    @stderr.setter
    def stderr(self, s):
        self._stderr = s
    
    @property
    def libcache(self) -> bool:
        """ Whether cache dll files. Only take effect in Windows emulation.

            Type: bool
            Example: - ql = Qiling(libcache=False)
                     - ql.libcache = True
        """
        return self._libcache

    @libcache.setter
    def libcache(self, lc):
        self._libcache = lc

    @property
    def output(self) -> int:
        """ Specify the qiling output. See Qiling.verbose.__doc__ for details.

            Note: Please pass None or one of the strings below to Qiling.__init__.

            Type: int
            Values:
              - "default": equals to "output=None", do nothing.
              - "off": an alias to "default".
              - "debug": set the log level to logging.DEBUG.
              - "disasm": diasm each executed instruction.
              - "dump": the most verbose output, dump registers and diasm the function blocks.
            Example: - ql = Qiling(output="off")
                     - ql.output = "off"
        """
        return self._output
    
    @output.setter
    def output(self, op):
        if type(op) is str:
            self._output = output_convert(op)
        else:
            self._output = op
        self.log.setLevel(ql_resolve_logger_level(self._output, self._verbose))
        self.os.setup_output()

    @property
    def verbose(self):
        """ Set the verbose level.
            
            If you set "ql.output" to "default" or "off", you can set logging level dynamically by
            changing "ql.verbose".

            Type: int
            Values:
              - 0  : logging.WARNING, almost no additional logs except the program output.
              - >=1: logging.INFO, the default logging level.
              - >=4: logging.DEBUG.
            Example: - ql = Qiling(verbose=5)
                     - ql.verbose = 0
        """
        return self._verbose
    
    @verbose.setter
    def verbose(self, v):
        self._verbose = v
        self.log.setLevel(ql_resolve_logger_level(self._output, self._verbose))
        self.os.setup_output()
    
    @property
    def patch_bin(self) -> list:
        """ Return the patches for binary.

            Type: list
        """
        return self._patch_bin
    
    @property
    def patch_lib(self) -> list:
        """ Return the patches for library.

            Type: list
        """
        return self._patch_lib

    @property
    def debug_stop(self) -> bool:
        """ Stop if some syscalls is not implemented.

            Note: This option is broken on some archs.

            Type: bool
            Example: ql.debug_stop = True
        """
        return self._debug_stop

    @debug_stop.setter
    def debug_stop(self, ds):
        self._debug_stop = ds

    @property
    def debugger(self) -> Union[str, bool]:
        """ Enable debugger.

            Type: debugger instance
            Values:
              - "gdb": enable gdb.
              - True : an alias to "gdb".
              - "gdb:0.0.0.0:1234" : gdb which listens on 0.0.0.0:1234
              - "qdb": enable qdb.
              - "qdb:rr": enable qdb with reverse debugging support.
            Example: ql.debugger = True
                     ql.debugger = "qdb"
        """
        return self._debugger
    
    @debugger.setter
    def debugger(self, dbger):
        self._debugger = dbger

    @property
    def root(self) -> bool:
        """ Whether run current program as root?

            Type: bool
            Examples: ql.root = True
        """
        return self._root

    @root.setter
    def root(self, root):
        self._root = root

    @property
    def filters(self) -> List[str]:
        """ Filter logs with regex.
            
            Type: List[str]
            Example: - Qiling(filters=[r'^exit'])
                     - ql.filters = [r'^open']
        """
        return self._filters

    @filters.setter
    def filters(self, ft):
        self._filters = ft
        if self._log_filter is None:
            self._log_filter = RegexFilter(ft)
            self.log.addFilter(self._log_filter)
        else:
            self._log_filter.update_filters(ft)

    @property
    def uc(self):
        """ Raw uc instance.

            Type: Uc
        """
        return self._uc

    @uc.setter
    def uc(self, u):
        self._uc = u

    @property
    def stop_options(self) -> "QlStopOptions":
        """ The stop options configured:
            - stackpointer: Stop execution on a negative stackpointer
            - exit_trap: Stop execution when the ip enters a guarded region
            - any: Is any of the options enabled?

        Returns:
            QlStopOptions: What stop options are configured
        """
        return self._stop_options

    def __enable_bin_patch(self):
        for addr, code in self.patch_bin:
            self.mem.write(self.loader.load_address + addr, code)


    def enable_lib_patch(self):
        for addr, code, filename in self.patch_lib:
            try:
                self.mem.write(self.mem.get_lib_base(filename) + addr, code)
            except:
                raise RuntimeError("Fail to patch %s at address 0x%x" % (filename, addr))

    def _init_stop_guard(self):
        if not self.stop_options.any:
            return

        # Allocate a guard page, we need this in both cases
        # On a negative stack pointer, we still need a return address (otherwise we end up at 0)
        # Make sure it is not close to the heap (PE), otherwise the heap cannot grow
        self._exit_trap_addr = self.mem.find_free_space(0x1000, min_addr=0x9000000, alignment=0x10)
        self.mem.map(self._exit_trap_addr, 0x1000, info='[Stop guard]')

        # Stop on a negative stack pointer
        if self.stop_options.stackpointer:
            def _check_sp(ql, address, size):
                if not ql.loader.skip_exit_check:
                    sp = ql._initial_sp - ql.reg.arch_sp
                    if sp < 0:
                        self.log.info('Process returned from entrypoint (stackpointer)!')
                        ql.emu_stop()

            self.hook_code(_check_sp)

        # Stop when running to exit trap address
        if self.stop_options.exit_trap:
            def _exit_trap(ql):
                self.log.info('Process returned from entrypoint (exit_trap)!')
                ql.emu_stop()

            self.hook_address(_exit_trap, self._exit_trap_addr)

    def write_exit_trap(self):
        self._initial_sp = self.reg.arch_sp
        if self.stop_options.any:
            if not self.loader.skip_exit_check:
                self.log.debug(f'Setting up exit trap at 0x{hex(self._exit_trap_addr)}')
                self.stack_write(0, self._exit_trap_addr)
            elif self.stop_options.exit_trap:
                self.log.debug(f'Loader {self.loader} requested to skip exit_trap!')


    ###############
    # Qiling APIS #
    ###############

    # Emulate the binary from begin until @end, with timeout in @timeout and
    # number of emulated instructions in @count
    def run(self, begin=None, end=None, timeout=0, count=0):
        # replace the original entry point, exit point, timeout and count
        self.entry_point = begin
        self.exit_point = end
        self.timeout = timeout
        self.count = count
        self.write_exit_trap()

        # init debugger
        if self._debugger != False and self._debugger != None:
            self._debugger = debugger_setup(self._debugger, self)

        # patch binary
        self.__enable_bin_patch()

        # emulate the binary
        self.os.run()

        # run debugger
        if self._debugger != False and self._debugger != None:
            self._debugger.run()


    # patch code to memory address
    def patch(self, addr, code, file_name=b''):
        if file_name == b'':
            self.patch_bin.append((addr, code))
        else:
            self.patch_lib.append((addr, code, file_name.decode()))


    # save all qiling instance states
    def save(self, reg=True, mem=True, fd=False, cpu_context=False, os_context=False, loader=False, snapshot=None):
        saved_states = {}

        if reg == True:
            saved_states.update({"reg": self.reg.save()})

        if mem == True:
            saved_states.update({"mem": self.mem.save()})

        if fd == True: 
            saved_states.update({"fd": self.os.fd.save()})

        if cpu_context == True:
            saved_states.update({"cpu_context": self.arch.context_save()})

        if os_context == True:
            saved_states.update({"os_context": self.os.save()})
        
        if loader == True:
            saved_states.update({"loader": self.loader.save()})

        if snapshot != None:
            with open(snapshot, "wb") as save_state:
                pickle.dump(saved_states, save_state)
        else:
            return saved_states


    # restore states qiling instance from saved_states
    def restore(self, saved_states=None, snapshot=None):

        # snapshot will be ignored if saved_states is set
        if saved_states == None and snapshot != None:
            with open(snapshot, "rb") as load_state:
                saved_states = pickle.load(load_state)

        if "cpu_context" in saved_states:
            self.arch.context_restore(saved_states["cpu_context"])

        if "reg" in saved_states:
            self.reg.restore(saved_states["reg"])

        if "mem" in saved_states:
            self.mem.restore(saved_states["mem"])
        
        if "fd" in saved_states:
            self.os.fd.restore(saved_states["fd"])

        if "os_context" in saved_states:
            self.os.restore(saved_states["os_context"])
        
        if "loader" in saved_states:
            self.loader.restore(saved_states["loader"])


    # Either hook or replace syscall/api with custom api/syscall
    #  - if intercept is None, replace syscall with custom function
    #  - if intercept is ENTER/EXIT, hook syscall at enter/exit with custom function
    # If replace function name is needed, first syscall must be available
    # - ql.set_syscall(0x04, my_syscall_write)
    # - ql.set_syscall("write", my_syscall_write)
    # TODO: Add correspoinding API in ql.os!
    def set_syscall(self, target_syscall, intercept_function, intercept = None):
        self.os.set_syscall(target_syscall, intercept_function, intercept)
        

    # Either replace or hook API
    #  - if intercept is None, replace API with custom function
    #  - if intercept is ENTER/EXIT, hook API at enter/exit with custom function
    def set_api(self, api_name, intercept_function, intercept = None):
        self.os.set_api(api_name, intercept_function, intercept)
 

    # Map "ql_path" to any objects which implements QlFsMappedObject.
    def add_fs_mapper(self, ql_path, real_dest):
        self.os.fs_mapper.add_fs_mapping(ql_path, real_dest)


    # push to stack bottom, and update stack register
    def stack_push(self, data):
        return self.arch.stack_push(data)


    # pop from stack bottom, and update stack register
    def stack_pop(self):
        return self.arch.stack_pop()


    # read from stack, at a given offset from stack bottom
    # NOTE: unlike stack_pop(), this does not change stack register
    def stack_read(self, offset):
        return self.arch.stack_read(offset)


    # write to stack, at a given offset from stack bottom
    # NOTE: unlike stack_push(), this does not change stack register
    def stack_write(self, offset, data):
        return self.arch.stack_write(offset, data)


    # Assembler/Diassembler API
    @property
    def assembler(self):
        return self.create_assembler()


    @property
    def disassembler(self):
        return self.create_disassembler()


    def create_disassembler(self):
        return self.arch.create_disassembler()


    def create_assembler(self):
        return self.arch.create_assembler()

    # stop emulation
    def emu_stop(self):
        self.uc.emu_stop()


    # start emulation
    def emu_start(self, begin, end, timeout=0, count=0):
        self.uc.emu_start(begin, end, timeout, count)
        
        if self._internal_exception != None:
            raise self._internal_exception



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from qiling.os.uefi import bs, rt, ds
from qiling.os.uefi.utils import install_configuration_table
from qiling.os.uefi.UefiSpec import EFI_SYSTEM_TABLE, EFI_BOOT_SERVICES, EFI_RUNTIME_SERVICES, EFI_CONFIGURATION_TABLE

# static mem layout:
#
#		+-- EFI_SYSTEM_TABLE ---------+
#		|                             |
#		| ...                         |
#		| RuntimeServices*     -> (1) |
#		| BootServices*        -> (2) |
#		| NumberOfTableEntries        |
#		| ConfigurationTable*  -> (4) |
#		+-----------------------------+
#	(1)	+-- EFI_RUNTIME_SERVICES -----+
#		|                             |
#		| ...                         |
#		+-----------------------------+
#	(2)	+-- EFI_BOOT_SERVICES --------+
#		|                             |
#		| ...                         |
#		+-----------------------------+
#	(3)	+-- EFI_DXE_SERVICES ---------+
#		|                             |
#		| ...                         |
#		+-----------------------------+
#	(4)	+-- EFI_CONFIGURATION_TABLE --+		of HOB_LIST
#		| VendorGuid                  |
#		| VendorTable*         -> (5) |
#		+-----------------------------+
#		+-- EFI_CONFIGURATION_TABLE --+		of DXE_SERVICE_TABLE
#		| VendorGuid                  |
#		| VendorTable*         -> (3) |
#		+-----------------------------+
#
#		... sizeof EFI_CONFIGURATION_TABLE x 98
#
#	(5)	+-- VOID* --------------------+
#		| ...                         |
#		+-----------------------------+
#
#		... the remainder of the 256 KiB chunk may be used for more conf table data

def initialize(ql, gST : int):
	ql.loader.gST = gST

	gBS = gST + EFI_SYSTEM_TABLE.sizeof()		# boot services
	gRT = gBS + EFI_BOOT_SERVICES.sizeof()		# runtime services
	gDS = gRT + EFI_RUNTIME_SERVICES.sizeof()	# dxe services
	cfg = gDS + ds.EFI_DXE_SERVICES.sizeof()	# configuration tables array

	ql.log.info(f'Global tables:')
	ql.log.info(f' | gST   {gST:#010x}')
	ql.log.info(f' | gBS   {gBS:#010x}')
	ql.log.info(f' | gRT   {gRT:#010x}')
	ql.log.info(f' | gDS   {gDS:#010x}')
	ql.log.info(f'')

	bs.initialize(ql, gBS)
	rt.initialize(ql, gRT)
	ds.initialize(ql, gDS)

	instance = EFI_SYSTEM_TABLE()
	instance.RuntimeServices = gRT
	instance.BootServices = gBS
	instance.NumberOfTableEntries = 0
	instance.ConfigurationTable = cfg

	instance.saveTo(ql, gST)

	# configuration tables bookkeeping
	confs = []

	# these are needed for utils.CoreInstallConfigurationTable
	ql.loader.dxe_context.conf_table_array = confs
	ql.loader.dxe_context.conf_table_array_ptr = cfg

	# configuration table data space; its location is calculated by leaving
	# enough space for 100 configuration table entries. only a few entries are
	# expected, so 100 should definitely suffice
	conf_data = cfg + EFI_CONFIGURATION_TABLE.sizeof() * 100
	ql.loader.dxe_context.conf_table_data_ptr = conf_data
	ql.loader.dxe_context.conf_table_data_next_ptr = conf_data

	install_configuration_table(ql.loader.dxe_context, "HOB_LIST", None)
	install_configuration_table(ql.loader.dxe_context, "DXE_SERVICE_TABLE", gDS)



__all__ = [
	'initialize'
]


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#
from Registry import Registry

# ERRORS CODE
# https://docs.microsoft.com/en-us/windows/win32/debug/system-error-codes--0-499-
ERROR_SUCCESS = 0x0
ERROR_INVALID_FUNCTION = 0x1
ERROR_FILE_NOT_FOUND = 0x2
ERROR_PATH_NOT_FOUND = 0x3
ERROR_ACCESS_DENIED = 0x5
ERROR_INVALID_HANDLE = 0x6
ERROR_INVALID_PARAMETER = 0x57
ERROR_INSUFFICIENT_BUFFER = 0x7A
ERROR_ALREADY_EXISTS = 0xB7
ERROR_MORE_DATA = 0xEA
ERROR_NOT_OWNER = 0x120
ERROR_OLD_WIN_VERSION = 0X47E
# ...

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-erref/596a1078-e883-4972-9bbc-49e60bebca55
STATUS_SUCCESS = 0
STATUS_UNSUCCESSFUL = 0xC0000001
# ...
STATUS_INFO_LENGTH_MISMATCH = 0xC0000004
STATUS_INVALID_PARAMETER = 0xC000000D
STATUS_INVALID_HANDLE = 0xC0000008
STATUS_PORT_NOT_SET = 0xC0000353
STATUS_NO_YIELD_PERFORMED = 0x40000024
# ...

INVALID_HANDLE_VALUE = -1

STD_INPUT_HANDLE = 0xfffffff6
STD_OUTPUT_HANDLE = 0xfffffff5
STD_ERROR_HANDLE = 0xfffffff4

# Registry Type

# Predefined Keys
REG_KEYS = {
    0x80000000: "HKEY_CLASSES_ROOT",
    0x80000005: "HKEY_CURRENT_CONFIG",
    0x80000001: "HKEY_CURRENT_USER",
    0x80000007: "HKEY_CURRENT_USER_LOCAL_SETTINGS",
    0x80000002: "HKEY_LOCAL_MACHINE",
    0x80000004: "HKEY_PERFORMANCE_DATA",
    0x80000060: "HKEY_PERFORMANCE_NLSTEXT",
    0x80000050: "HKEY_PERFORMANCE_TEXT",
    0x80000003: "HKEY_USERS"
}

REG_TYPES = {
    "REG_NONE": Registry.RegNone,
    "REG_SZ": Registry.RegSZ,
    "REG_EXPAND_SZ": Registry.RegExpandSZ,
    "REG_BINARY": Registry.RegBin,
    "REG_DWORD": Registry.RegDWord,
    "REG_DWORD_BIG_ENDIAN": Registry.RegBigEndian,
    "REG_LINK": Registry.RegLink,
    "REG_MULTI_SZ": Registry.RegMultiSZ,
    "REG_RESOURCE_LIST": Registry.RegResourceList,
    "REG_FULL_RESOURCE_DESCRIPTOR": Registry.RegFullResourceDescriptor,
    "REG_RESOURCE_REQUIREMENTS_LIST": Registry.RegResourceRequirementsList,
    "REG_QWORD": Registry.RegQWord,

    Registry.RegNone: "REG_NONE",
    Registry.RegSZ: "REG_SZ",
    Registry.RegExpandSZ: "REG_EXPAND_SZ",
    Registry.RegBin: "REG_BINARY",
    Registry.RegDWord: "REG_DWORD",
    Registry.RegBigEndian: "REG_DWORD_BIG_ENDIAN",
    Registry.RegLink: "REG_LINK",
    Registry.RegMultiSZ: "REG_MULTI_SZ",
    Registry.RegResourceList: "REG_RESOURCE_LIST",
    Registry.RegFullResourceDescriptor: "REG_FULL_RESOURCE_DESCRIPTOR",
    Registry.RegResourceRequirementsList: "REG_RESOURCE_REQUIREMENTS_LIST",
    Registry.RegQWord: "REG_QWORD"
}

RTL_MAXIMUM_ATOM_LENGTH = 255
RTL_USER_PROCESS_PARAMETERS_NORMALIZED = 0x01
RTL_USER_PROCESS_PARAMETERS_PROFILE_USER = 0x02
RTL_USER_PROCESS_PARAMETERS_PROFILE_KERNEL = 0x04
RTL_USER_PROCESS_PARAMETERS_PROFILE_SERVER = 0x08
RTL_USER_PROCESS_PARAMETERS_UNKNOWN = 0x10
RTL_USER_PROCESS_PARAMETERS_RESERVE_1MB = 0x20
RTL_USER_PROCESS_PARAMETERS_RESERVE_16MB = 0x40
RTL_USER_PROCESS_PARAMETERS_CASE_SENSITIVE = 0x80
RTL_USER_PROCESS_PARAMETERS_DISABLE_HEAP_CHECKS = 0x100
RTL_USER_PROCESS_PARAMETERS_PROCESS_OR_1 = 0x200
RTL_USER_PROCESS_PARAMETERS_PROCESS_OR_2 = 0x400
RTL_USER_PROCESS_PARAMETERS_PRIVATE_DLL_PATH = 0x1000
RTL_USER_PROCESS_PARAMETERS_LOCAL_DLL_PATH = 0x2000
RTL_USER_PROCESS_PARAMETERS_IMAGE_KEY_MISSING = 0x4000
RTL_USER_PROCESS_PARAMETERS_NX = 0x20000
RTL_MAX_DRIVE_LETTERS = 32
RTL_DRIVE_LETTER_VALID = 0x0001
# EXCEPTION_CHAIN_END = ((PEXCEPTION_REGISTRATION_RECORD)-1)
SEM_FAILCRITICALERRORS = 0x0001
SEM_NOGPFAULTERRORBOX = 0x0002
SEM_NOALIGNMENTFAULTEXCEPT = 0x0004
SEM_NOOPENFILEERRORBOX = 0x8000
RTL_SEM_FAILCRITICALERRORS = (SEM_FAILCRITICALERRORS << 4)
RTL_SEM_NOGPFAULTERRORBOX = (SEM_NOGPFAULTERRORBOX << 4)
RTL_SEM_NOALIGNMENTFAULTEXCEPT = (SEM_NOALIGNMENTFAULTEXCEPT << 4)
RTL_RANGE_LIST_ADD_IF_CONFLICT = 0x00000001
RTL_RANGE_LIST_ADD_SHARED = 0x00000002
RTL_RANGE_SHARED = 0x01
RTL_RANGE_CONFLICT = 0x02
RTL_ACTIVATION_CONTEXT_STACK_FRAME_FLAG_RELEASE_ON_DEACTIVATION = 0x01
RTL_ACTIVATION_CONTEXT_STACK_FRAME_FLAG_NO_DEACTIVATE = 0x02
RTL_ACTIVATION_CONTEXT_STACK_FRAME_FLAG_ON_FREE_LIST = 0x04
RTL_ACTIVATION_CONTEXT_STACK_FRAME_FLAG_HEAP_ALLOCATED = 0x08
RTL_ACTIVATION_CONTEXT_STACK_FRAME_FLAG_NOT_REALLY_ACTIVATED = 0x10
RTL_ACTIVATION_CONTEXT_STACK_FRAME_FLAG_ACTIVATED = 0x20
RTL_ACTIVATION_CONTEXT_STACK_FRAME_FLAG_DEACTIVATED = 0x40
RTL_CALLER_ALLOCATED_ACTIVATION_CONTEXT_STACK_FRAME_FORMAT_WHISTLER = 0x01
RTL_ACTIVATE_ACTIVATION_CONTEXT_EX_FLAG_RELEASE_ON_STACK_DEALLOCATION = 0x01
RTL_DEACTIVATE_ACTIVATION_CONTEXT_FLAG_FORCE_EARLY_DEACTIVATION = 0x01
RTL_QUERY_ACTIVATION_CONTEXT_FLAG_USE_ACTIVE_ACTIVATION_CONTEXT = 0x01
RTL_QUERY_ACTIVATION_CONTEXT_FLAG_IS_HMODULE = 0x02
RTL_QUERY_ACTIVATION_CONTEXT_FLAG_IS_ADDRESS = 0x04
RTL_QUERY_ACTIVATION_CONTEXT_FLAG_NO_ADDREF = 0x80000000
HEAP_SETTABLE_USER_VALUE = 0x00000100
HEAP_SETTABLE_USER_FLAG1 = 0x00000200
HEAP_SETTABLE_USER_FLAG2 = 0x00000400
HEAP_SETTABLE_USER_FLAG3 = 0x00000800
HEAP_SETTABLE_USER_FLAGS = 0x00000E00
HEAP_CLASS_0 = 0x00000000
HEAP_CLASS_1 = 0x00001000
HEAP_CLASS_2 = 0x00002000
HEAP_CLASS_3 = 0x00003000
HEAP_CLASS_4 = 0x00004000
HEAP_CLASS_5 = 0x00005000
HEAP_CLASS_6 = 0x00006000
HEAP_CLASS_7 = 0x00007000
HEAP_CLASS_8 = 0x00008000
HEAP_CLASS_MASK = 0x0000F000
HEAP_FLAG_ALLOCS = 0x01000000
HEAP_PROTECTION_ENABLED = 0x02000000
HEAP_BREAK_WHEN_OUT_OF_VM = 0x04000000
HEAP_NO_ALIGNMENT = 0x08000000
HEAP_CAPTURE_STACK_BACKTRACES = 0x08000000
HEAP_SKIP_VALIDATION_CHECKS = 0x10000000
HEAP_VALIDATE_ALL_ENABLED = 0x20000000
HEAP_VALIDATE_PARAMETERS_ENABLED = 0x40000000
HEAP_LOCK_USER_ALLOCATED = 0x80000000
# HEAP_CREATE_VALID_MASK
RTL_REGISTRY_ABSOLUTE = 0
RTL_REGISTRY_SERVICES = 1
RTL_REGISTRY_CONTROL = 2
RTL_REGISTRY_WINDOWS_NT = 3
RTL_REGISTRY_DEVICEMAP = 4
RTL_REGISTRY_USER = 5
RTL_REGISTRY_MAXIMUM = 6
RTL_REGISTRY_HANDLE = 0x40000000
RTL_REGISTRY_OPTIONAL = 0x80000000
RTL_QUERY_REGISTRY_SUBKEY = 0x00000001
RTL_QUERY_REGISTRY_TOPKEY = 0x00000002
RTL_QUERY_REGISTRY_REQUIRED = 0x00000004
RTL_QUERY_REGISTRY_NOVALUE = 0x00000008
RTL_QUERY_REGISTRY_NOEXPAND = 0x00000010
RTL_QUERY_REGISTRY_DIRECT = 0x00000020
RTL_QUERY_REGISTRY_DELETE = 0x00000040
VER_MINORVERSION = 0x0000001
VER_MAJORVERSION = 0x0000002
VER_BUILDNUMBER = 0x0000004
VER_PLATFORMID = 0x0000008
VER_SERVICEPACKMINOR = 0x0000010
VER_SERVICEPACKMAJOR = 0x0000020
VER_SUITENAME = 0x0000040
VER_PRODUCT_TYPE = 0x0000080
VER_PLATFORM_WIN32s = 0
VER_PLATFORM_WIN32_WINDOWS = 1
VER_PLATFORM_WIN32_NT = 2
VER_EQUAL = 1
VER_GREATER = 2
VER_GREATER_EQUAL = 3
VER_LESS = 4
VER_LESS_EQUAL = 5
VER_AND = 6
VER_OR = 7
VER_CONDITION_MASK = 7
VER_NUM_BITS_PER_CONDITION_MASK = 3
TIME_ZONE_ID_UNKNOWN = 0
TIME_ZONE_ID_STANDARD = 1
TIME_ZONE_ID_DAYLIGHT = 2
MAX_PATH = 260
RTL_CRITSECT_TYPE = 0
RTL_RESOURCE_TYPE = 1
RTL_ACQUIRE_PRIVILEGE_IMPERSONATE = 1
RTL_ACQUIRE_PRIVILEGE_PROCESS = 2
MESSAGE_RESOURCE_UNICODE = 0x0001
MAXIMUM_LEADBYTES = 12
RTL_DEBUG_QUERY_MODULES = 0x01
RTL_DEBUG_QUERY_BACKTRACES = 0x02
RTL_DEBUG_QUERY_HEAPS = 0x04
RTL_DEBUG_QUERY_HEAP_TAGS = 0x08
RTL_DEBUG_QUERY_HEAP_BLOCKS = 0x10
RTL_DEBUG_QUERY_LOCKS = 0x20
RTL_HANDLE_VALID = 0x1
RTL_ATOM_IS_PINNED = 0x1
CS_LOCK_BIT = 0x1
CS_LOCK_BIT_V = 0x0
CS_LOCK_WAITER_WOKEN = 0x2
CS_LOCK_WAITER_INC = 0x4
# RTL_CONSTANT_LARGE_INTEGER(quad_part) = { { (quad_part), (quad_part)>>32 } }
# RTL_MAKE_LARGE_INTEGER(low_part, high_part) = { { (low_part), (high_part) } }
RTL_FLS_MAXIMUM_AVAILABLE = 128
RTL_RESOURCE_FLAG_LONG_TERM = 0x00000001
# UMSCTX_SCHEDULED_THREAD_MASK = (1 << UMSCTX_SCHEDULED_THREAD_BIT)
# UMSCTX_SUSPENDED_MASK = (1 << UMSCTX_SUSPENDED_BIT)
# UMSCTX_VOLATILE_CONTEXT_MASK = (1 << UMSCTX_VOLATILE_CONTEXT_BIT)
# UMSCTX_TERMINATED_MASK = (1 << UMSCTX_TERMINATED_BIT)
# UMSCTX_DEBUG_ACTIVE_MASK = (1 << UMSCTX_DEBUG_ACTIVE_BIT)
# UMSCTX_RUNNING_ON_SELF_THREAD_MASK = (1 << UMSCTX_RUNNING_ON_SELF_THREAD_BIT)
# UMSCTX_DENY_RUNNING_ON_SELF_THREAD_MASK = (1 << UMSCTX_DENY_RUNNING_ON_SELF_THREAD_BIT)


# File operation
GENERIC_ALL = 0x10000000
GENERIC_WRITE = 0x40000000
GENERIC_READ = 0x80000000
GENERIC_EXECUTE = 0x20000000

CREATE_NEW = 1
CREATE_ALWAYS = 2
OPEN_EXISTING = 3
OPEN_ALWAYS = 4
TRUNCATE_EXISTING = 5

# HRESULT values
S_OK = 0x00
S_FALSE = 0x01
E_ABORT = 0x80004004
E_ACCESSDENIED = 0x80070005
E_FAIL = 0x80004005
E_HANDLE = 0x80070006
E_INVALIDARG = 0x80070057
E_NOINTERFACE = 0x80004002
E_NOTIMPL = 0x80004001
E_OUTOFMEMORY = 0x8007000E
E_POINTER = 0x80004003
E_UNEXPECTED = 0x8000FFFF

# LOCALE values
# https://docs.microsoft.com/it-it/windows/win32/intl/locale-custom-constants
LOCALE_USER_DEFAULT = 0x0400
LOCALE_CUSTOM_DEFAULT = 0x0C00
LOCALE_CUSTOM_UI_DEFAULT = 0x1400
LOCALE_CUSTOM_UNSPECIFIED = 0x1000

LOCALE_EN_US = {
    # http://www.borgendale.com/locale/en_US.htm
    0x1004: "utf-8",
    0x1005: "\x00",
    0x1009: "1",
    "sName": "en_US",
    "xLocaleToken": "100210e9",
    "xWinLocale": "0409",
    "sEngLanguage": "English",
    "sEngCountry": "United States",
    "sCountry": "United States",
    "sLanguage": "English",
    "sNativeCtryName": "United States",
    "sLanguageID": "en",
    "sCountryID": "US",
    "iCountry": "1",
    "sAbbrevLangName": "ENU",
    "jISO3CountryName": "USA",
    "iCodepage": "437",
    "iAltCodepage": "850",
    "iAnsiCodepage": "1252",
    "sISOCodepage": "ISO8859-1",
    "iMacCodepage": "1275",
    "iEbcdicCodepage": "37",
    "sSetCodepage": "",
    "sKeyboard": "us",
    "sCollate": "",
    "sCurrency": "$",
    "sIntlSymbol": "USD",
    "iCurrency": 0,
    "iCurrDigits": 2,
    "iNegCurr": 0,
    "sMonDecimalSep": ".",
    "sMonThousandSep": ",",
    "sMonGrouping": "«3»",
    "iDigits": 2,
    "iNegNumber": 1,
    "sDecimal": ".",
    "sThousand": ",",
    "sGrouping": "«3»",
    "iLzero": 0,
    "iTLzero": 0,
    "sList": ",",
    "iMeasure": 1,
    "iPaper": 1,
    "iTime": 0,
    "iDate": 0,
    "sDate": "/",
    "sTime": ":",
    "sDateTime": "%a %b %e %H:%M:%S %Z %Y",
    "sShortDate": "%m/%d/%y",
    "sTimeFormat": "%I:%M:%S %p",
    "sLongDate": "%B %d, %Y",
    "wTimeFormat": "hh:mm:ss tt",
    "wShortDate": "MM/dd/yy",
    "wLongDate": "MMMM dd, yyyy",
    "iLDate": 0,
    "iCalendarType": 1,
    "iFirstDayOfWeek": 6,
    "iFirstWeekOfYear": 0,
    "iTimePrefix": 0,
    "s1159": "am",
    "s2359": "pm",
    "sDayName7": "Sunday",
    "sDayName1": "Monday",
    "sDayName2": "Tuesday",
    "sDayName3": "Wednesday",
    "sDayName4": "Thursday",
    "sDayName5": "Friday",
    "sDayName6": "Saturday",
    "sAbbrevDayName7": "Sun",
    "sAbbrevDayName1": "Mon",
    "sAbbrevDayName2": "Tue",
    "sAbbrevDayName3": "Wed",
    "sAbbrevDayName4": "Thu",
    "sAbbrevDayName5": "Fri",
    "sAbbrevDayName6": "Sat",
    "sMonthName1": "January",
    "sMonthName2": "February",
    "sMonthName3": "March",
    "sMonthName4": "April",
    "sMonthName5": "May",
    "sMonthName6": "June",
    "sMonthName7": "July",
    "sMonthName8": "August",
    "sMonthName9": "September",
    "sMonthName10": "October",
    "sMonthName11": "November",
    "sMonthName12": "December",
    "sAbbrevMonthName1": "Jan",
    "sAbbrevMonthName2": "Feb",
    "sAbbrevMonthName3": "Mar",
    "sAbbrevMonthName4": "Apr",
    "sAbbrevMonthName5": "May",
    "sAbbrevMonthName6": "Jun",
    "sAbbrevMonthName7": "Jul",
    "sAbbrevMonthName8": "Aug",
    "sAbbrevMonthName9": "Sep",
    "sAbbrevMonthName10": "Oct",
    "sAbbrevMonthName11": "Nov",
    "sAbbrevMonthName12": "Dec",
    "sYesString": "yes:y:Y",
    "sNoString": "no:n:N",
    "sNativeDigits": "0123456789",
    "sYesExpr": "^([yY]|[yY][eE][sS])",
    "sNoExpr": "^([nN]|[nN][oO])",
    "iUpperType": 0,
    "iUpperMissing": 0,
    "sPositiveSign": "",
    "sNegativeSign": "-",
    "sLeftNegative": "(",
    "sRightNegative": ")",
    "sDebit": "DB",
    "sCredit": "CR",
    "jPercentPattern": "#,##0%",
    "jPercentSign": "%",
    "jExponent": "E",
    "jFullTimeFormat": "HH:mm:ss 'o''''clock' z",
    "jLongTimeFormat": "HH:mm:ss z",
    "jShortTimeFormat": "HH:mm",
    "jFullDateFormat": "EEEE, MMMM d, yyyy",
    "jMediumDateFormat": "dd-MMM-yy",
    "jDateTimePattern": "{1} {0}"
}
LOCALE = {
    0x409: LOCALE_EN_US,
    "default": LOCALE_EN_US
}
# SYSTEM_INFORMATION_CLASS
# Defined in Winternl.h
# Used by NTQuerySystemInformation and ZwQuerySystemInformation
SystemBasicInformation = 0
SystemPerformanceInformation = 2
SystemTimeOfDayInformation = 3
SystemProcessInformation = 5
SystemProcessorPerformanceInformation = 8
SystemInterruptInformation = 23
SystemExceptionInformation = 33
SystemRegistryQuotaInformation = 37
SystemLookasideInformation = 45
# Code Page Identifiers
# https://docs.microsoft.com/en-us/windows/win32/intl/code-page-identifiers
OEM_US = 437

# SystemMetrics
# https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getsystemmetrics
SM_ARRANGE = 56
SM_CLEANBOOT = 67
SM_CMONITORS = 80
SM_CMOUSEBUTTONS = 43
SM_CONVERTIBLESLATEMODE = 0x2003
SM_CXBORDER = 5
SM_CXCURSOR = 13
SM_CXDLGFRAME = 7
SM_CXDOUBLECLK = 36
SM_CXDRAG = 68
SM_CXEDGE = 45
SM_CXFIXEDFRAME = 7
SM_CXFOCUSBORDER = 83
SM_CXFRAME = 32
SM_CXFULLSCREEN = 16
SM_CXHSCROLL = 21
SM_CXHTHUMB = 10
SM_CXICON = 11
SM_CXICONSPACING = 38
SM_CXMAXIMIZED = 61
SM_CXMAXTRACK = 59
SM_CXMENUCHECK = 71
SM_CXMENUSIZE = 54
SM_CXMIN = 28
SM_CXMINIMIZED = 57
SM_CXMINSPACING = 47
SM_CXMINTRACK = 34
SM_CXPADDEDBORDER = 92
SM_CXSCREEN = 0
SM_CXSIZE = 30
SM_CXSIZEFRAME = 32
SM_CXSMICON = 49
SM_CXSMSIZE = 52
SM_CXVIRTUALSCREEN = 78
SM_CXVSCROLL = 2
SM_CYBORDER = 6
SM_CYCAPTION = 4
SM_CYCURSOR = 14
SM_CYDLGFRAME = 8
SM_CYDOUBLECLK = 37
SM_CYDRAG = 69
SM_CYEDGE = 46
SM_CYFIXEDFRAME = 8
SM_CYFOCUSBORDER = 84
SM_CYFRAME = 33
SM_CYFULLSCREEN = 17
SM_CYHSCROLL = 3
SM_CYICON = 12
# Got bored, will add more if is necessary


# VirtualKey Mappings
# https://docs.microsoft.com/it-it/windows/win32/inputdev/virtual-key-codes
# https://www.win.tue.nl/~aeb/linux/kbd/scancodes-1.html
MAPVK_VK_TO_VSC = {
    0x5b: 0x5b,
    0x10: 0x2a,
    0xa0: 0x2a,
    0xa1: 0x36,
    0x11: 0x1d,
    0x12: 0x38
}
# more to add, just inserted the necessary for a sample

MAP_VK = {
    0: MAPVK_VK_TO_VSC
}

# Crypt String Mappings
# https://docs.microsoft.com/en-us/windows/win32/api/wincrypt/nf-wincrypt-cryptstringtobinarya

CRYPT_STRING_BASE64HEADER = 0
CRYPT_STRING_BASE64 = 1
CRYPT_STRING_BINARY = 2
CRYPT_STRING_BASE64REQUESTHEADER = 3
# ...

# File Attribtues Constantsc
# https://docs.microsoft.com/en-us/windows/win32/fileio/file-attribute-constants
FILE_ATTRIBUTE_ARCHIVE = 0x0020
FILE_ATTRIBUTE_COMPRESSED = 0x0800
FILE_ATTRIBUTE_DEVICE = 0x0040
FILE_ATTRIBUTE_DIRECTORY = 0x0010
FILE_ATTRIBUTE_ENCRYPTED = 0x4000
FILE_ATTRIBUTE_HIDDEN = 0x0002
FILE_ATTRIBUTE_INTEGRITY_STREAM = 0x8000
FILE_ATTRIBUTE_NORMAL = 0x0080
FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = 0x2000
FILE_ATTRIBUTE_NO_SCRUB_DATA = 0x20000
FILE_ATTRIBUTE_OFFLINE = 0x1000
FILE_ATTRIBUTE_READONLY = 0x0001
FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS = 0x400000
FILE_ATTRIBUTE_RECALL_ON_OPEN = 0x40000
FILE_ATTRIBUTE_REPARSE_POINT = 0x400
FILE_ATTRIBUTE_SPARSE_FILE = 0x0004
FILE_ATTRIBUTE_TEMPORARY = 0x0100
FILE_ATTRIBUTE_VIRTUAL = 0x10000

# ShellApi Constants
# https://docs.microsoft.com/en-us/windows/win32/api/shellapi/nf-shellapi-shgetfileinfow

SHGFI_ADDOVERLAYS = 0x000000020
SHGFI_ATTR_SPECIFIED = 0x000020000
SHGFI_ATTRIBUTES = 0x000000800
SHGFI_DISPLAYNAME = 0x000000200
SHGFI_EXETYPE = 0x000002000
SHGFI_ICON = 0x000000100
SHGFI_ICONLOCATION = 0x000001000
SHGFI_LARGEICON = 0x000000000
# ...

# Tlhelp32 Constans
# https://docs.microsoft.com/en-us/windows/win32/api/tlhelp32/nf-tlhelp32-createtoolhelp32snapshot
TH32CS_INHERIT = 0x80000000
TH32CS_SNAPHEAPLIST = 0x00000001
TH32CS_SNAPMODULE = 0x00000008
TH32CS_SNAPMODULE32 = 0x00000010
TH32CS_SNAPPROCESS = 0x00000002
TH32CS_SNAPTHREAD = 0x00000004

# CSIDL Constants
# https://tarma.com/support/im9/using/symbols/functions/csidls.htm
# http://winbatch.hpdd.de/MyWbtHelp/other/CSIDL.txt
CSIDL_ADMINTOOLS = 0x30
CSIDL_ALTSTARTUP = 0x1D
CSIDL_APPDATA = 0x1A
CSIDL_BITBUCKET = 0x0A
CSIDL_CDBURN_AREA = 0x3B
CSIDL_COMMON_ADMINTOOLS = 0x2F
CSIDL_COMMON_ALTSTARTUP = 0x1E
CSIDL_COMMON_APPDATA = 0x23
# ...
CSIDL_FLAG_CREATE = 0x8000

# Show constants
# https://docs.microsoft.com/it-it/windows/win32/api/shellapi/nf-shellapi-shellexecutea
SW_HIDE = 0
SW_MAXIMIZE = 3
SW_MINIMIZE = 6
SW_RESTORE = 9
SW_SHOW = 5
SW_SHOWDEFAULT = 10
SW_SHOWMAXIMIZED = 3
SW_SHOWMINIMIZED = 2
SW_SHOWMINNOACTIVE = 7
SW_SHOWNA = 8
SW_SHOWNOACTIVATE = 4
SW_SHOWNORMAL = 1

# OsVersionInfoConstants
# https://docs.microsoft.com/it-it/windows/win32/api/winnt/ns-winnt-osversioninfoexa

VER_NT_DOMAIN_CONTROLLER = 0x0000002
VER_NT_SERVER = 0x0000003
VER_NT_WORKSTATION = 0x0000001

# major, minor, product
SYSTEMS_VERSION = {
    "1001": "Windows 10",
    "1000": "Windows Server 2016",

    "631": "Windows 8.1",
    "630": "Windows Server 2012 R2",

    "621": "Windows 8",
    "620": "Windows Server 2012",

    "611": "Windows 7",
    "610": "Windows Server 2008 R2",

    "601": "Windows Vista",
    "600": "Windows Server 2008",

    # ...

    "510": "Windows XP"
    # ...
}

# Mapper for ordinal syscalls
Mapper = {
    "shell32": {
        175: "SHGetSpecialFolderPathW"
    }
}

MAXUSHORT = 0xffff

DRIVE_UNKNOWN = 0
DRIVE_NO_ROOT_DIR = 1
DRIVE_REMOVABLE = 2
DRIVE_FIXED = 3
DRIVE_REMOTE = 4
DRIVE_CDROM = 5
DRIVE_RAMDISK = 6

# https://docs.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-getfiletype
FILE_TYPE_CHAR = 0x2
FILE_TYPE_DISK = 0x1
FILE_TYPE_PIPE = 0x3
FILE_TYPE_REMOTE = 0x800
FILE_TYPE_UNKNOWN = 0x0

# https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-isprocessorfeaturepresent
PF_XSAVE_ENABLED = 0x17
# ...


# https://docs.microsoft.com/en-us/windows/win32/procthread/zwqueryinformationprocess
ProcessBasicInformation = 0
ProcessDebugPort = 7
ProcessExecuteFlags = 0x22
ProcessWow64Information = 26
ProcessImageFileName = 27
ProcessBreakOnTermination = 29
ProcessProtectionInformation = 61
ProcessDebugObjectHandle = 0x1E
ProcessDebugFlags = 0x1F

# https://www.geoffchappell.com/studies/windows/km/ntoskrnl/api/ps/psquery/class.htm
ThreadBasicInformation = 0x0
ThreadTimes = 0x1
ThreadPriority = 0x2
# ...
ThreadHideFromDebugger = 0x11
# ...


# https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-messageboxw

MB_ABORTRETRYIGNORE = 0x000000020
MB_CANCELTRYCONTINUE = 0x000000060
MB_HELP = 0x000040000
MB_OK = 0x000000000
MB_OKCANCEL = 0x000000010
MB_RETRYCANCEL = 0x000000050
MB_YESNO = 0x000000040
MB_YESNOCANCEL = 0x000000030

IDABORT = 3
IDCANCEL = 2
IDCONTINUE = 11
IDIGNORE = 5
IDNO = 7
IDOK = 1
IDRETRY = 4
IDTRYAGAIN = 10
IDYES = 6

# https://docs.microsoft.com/en-us/windows/win32/api/libloaderapi/nf-libloaderapi-setdefaultdlldirectories

LOAD_LIBRARY_SEARCH_APPLICATION_DIR = 0x00000200
LOAD_LIBRARY_SEARCH_DEFAULT_DIRS = 0x00001000
LOAD_LIBRARY_SEARCH_SYSTEM32 = 0x00000800
LOAD_LIBRARY_SEARCH_USER_DIRS = 0x00000400

# https://groups.google.com/forum/#!topic/microsoft.public.vc.language/UERvWrmTn-o
CSTR_LESS_THAN = 1
CSTR_EQUAL = 2
CSTR_GREATER_THAN = 3

# https://docs.microsoft.com/it-it/windows/win32/memory/memory-protection-constants
PAGE_EXECUTE = 0x10
PAGE_EXECUTE_READ = 0x20
PAGE_EXECUTE_READWRITE = 0x40
PAGE_EXECUTE_WRITECOPY = 0x80
PAGE_NOACCESS = 0x1
PAGE_READONLY = 0x2
PAGE_READWRITE = 0x4
PAGE_WRITECOPY = 0x8
PAGE_TARGETS_INVALID = 0x40000000
PAGE_TARGETS_NO_UPDATE = 0x40000000
PAGE_GUARD = 0x100
PAGE_NOCACHE = 0x200
PAGE_WRITECOMBINE = 0x400


# https://docs.microsoft.com/en-us/windows/win32/api/winternl/nf-winternl-ntqueryobject
# https://gist.github.com/soxfmr/16c495d6e4ad99e9e46f5bfd558d152f
ObjectTypeInformation = 0x2
ObjectBasicInformation = 0x1
ObjectAllTypesInformation = 0x3

# https://docs.microsoft.com/en-us/windows/win32/api/handleapi/nf-handleapi-sethandleinformation
HANDLE_FLAG_INHERIT = 0x1
HANDLE_FLAG_PROTECT_FROM_CLOSE = 0x2



#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *
from qiling.os.stat import *


def create_stat_struct(ql, info):
    if ql.archtype == QL_ARCH.MIPS:
        # pack fstatinfo
        stat_buf = ql.pack32(info.st_dev)
        stat_buf += ql.pack32(0) * 3
        stat_buf += ql.pack32(info.st_ino)
        stat_buf += ql.pack32(info.st_mode)
        stat_buf += ql.pack32(info.st_nlink)
        stat_buf += ql.pack32(info.st_uid)
        stat_buf += ql.pack32(info.st_gid)
        stat_buf += ql.pack32(info.st_rdev)
        stat_buf += ql.pack32(0) * 2
        stat_buf += ql.pack32(info.st_size)
        stat_buf += ql.pack32(0)
        stat_buf += ql.pack32(int(info.st_atime))
        stat_buf += ql.pack32(0)
        stat_buf += ql.pack32(int(info.st_mtime))
        stat_buf += ql.pack32(0)
        stat_buf += ql.pack32(int(info.st_ctime))
        stat_buf += ql.pack32(0)
        stat_buf += ql.pack32(info.st_blksize)
        stat_buf += ql.pack32(info.st_blocks)
        stat_buf = stat_buf.ljust(0x90, b'\x00')
    elif ql.archtype == QL_ARCH.X8664:
        if ql.platform == QL_OS.MACOS:
            stat_buf = ql.pack64s(info.st_dev)
        else:
            stat_buf = ql.pack64(info.st_dev)
        stat_buf += ql.pack(info.st_ino)
        stat_buf += ql.pack64(info.st_nlink)
        stat_buf += ql.pack32(info.st_mode)
        stat_buf += ql.pack32(info.st_uid)
        stat_buf += ql.pack32(info.st_gid)
        stat_buf += ql.pack32(0)
        stat_buf += ql.pack64(info.st_rdev)
        stat_buf += ql.pack64(info.st_size)
        stat_buf += ql.pack64(info.st_blksize)
        stat_buf += ql.pack64(info.st_blocks)
        stat_buf += ql.pack64(int(info.st_atime))
        stat_buf += ql.pack64(0)
        stat_buf += ql.pack64(int(info.st_mtime))
        stat_buf += ql.pack64(0)
        stat_buf += ql.pack64(int(info.st_ctime))
        stat_buf += ql.pack64(0)
    elif ql.archtype == QL_ARCH.ARM64:
        # struct stat is : 80 addr is : 0x4000811bc8
        # buf.st_dev offest 0 8 0
        # buf.st_ino offest 8 8 0
        # buf.st_mode offest 10 4 0
        # buf.st_nlink offest 14 4 0
        # buf.st_uid offest 18 4 0
        # buf.st_gid offest 1c 4 0
        # buf.st_rdev offest 20 8 0
        # buf.st_size offest 30 8 274886889936
        # buf.st_blksize offest 38 4 8461328
        # buf.st_blocks offest 40 8 274877909532
        # buf.st_atime offest 48 8 274886368336
        # buf.st_mtime offest 58 8 274877909472
        # buf.st_ctime offest 68 8 274886368336
        # buf.__glibc_reserved offest 78 8
        if ql.platform == QL_OS.MACOS:
            stat_buf = ql.pack64s(info.st_dev)
        else:
            stat_buf = ql.pack64(info.st_dev)
        stat_buf += ql.pack64(info.st_ino)
        stat_buf += ql.pack32(info.st_mode)
        stat_buf += ql.pack32(info.st_nlink)
        stat_buf += ql.pack32(ql.os.uid)
        stat_buf += ql.pack32(ql.os.gid)
        stat_buf += ql.pack64(info.st_rdev)
        stat_buf += ql.pack64(0)
        stat_buf += ql.pack64(info.st_size)
        stat_buf += ql.pack32(info.st_blksize)
        stat_buf += ql.pack32(0)
        stat_buf += ql.pack64(info.st_blocks)
        stat_buf += ql.pack64(int(info.st_atime))
        stat_buf += ql.pack64(0)
        stat_buf += ql.pack64(int(info.st_mtime))
        stat_buf += ql.pack64(0)
        stat_buf += ql.pack64(int(info.st_ctime))
        stat_buf += ql.pack64(0)
    else:
        # pack fstatinfo
        stat_buf = ql.pack32(info.st_dev)
        stat_buf += ql.pack(info.st_ino)
        stat_buf += ql.pack32(info.st_mode)
        stat_buf += ql.pack32(info.st_nlink)
        stat_buf += ql.pack32(info.st_uid)
        stat_buf += ql.pack32(info.st_gid)
        stat_buf += ql.pack32(info.st_rdev)
        stat_buf += ql.pack32(info.st_size)
        stat_buf += ql.pack32(info.st_blksize)
        stat_buf += ql.pack32(info.st_blocks)
        stat_buf += ql.pack32(int(info.st_atime))
        stat_buf += ql.pack32(int(info.st_mtime))
        stat_buf += ql.pack32(int(info.st_ctime))
    return stat_buf


def create_stat64_struct(ql, info):
    if ql.archtype == QL_ARCH.ARM64:
        # struct stat is : 80 addr is : 0x4000811bc8
        # buf.st_dev offest 0 8 0
        # buf.st_ino offest 8 8 0
        # buf.st_mode offest 10 4 0
        # buf.st_nlink offest 14 4 0
        # buf.st_uid offest 18 4 0
        # buf.st_gid offest 1c 4 0
        # buf.st_rdev offest 20 8 0
        # buf.st_size offest 30 8 274886889936
        # buf.st_blksize offest 38 4 8461328
        # buf.st_blocks offest 40 8 274877909532
        # buf.st_atime offest 48 8 274886368336
        # buf.st_mtime offest 58 8 274877909472
        # buf.st_ctime offest 68 8 274886368336
        # buf.__glibc_reserved offest 78 8
        fstat64_buf = ql.pack64(info.st_dev)  # 8
        fstat64_buf += ql.pack64(info.st_ino)  # 16
        fstat64_buf += ql.pack32(info.st_mode)  # 20
        fstat64_buf += ql.pack32(info.st_nlink)  # 24
        fstat64_buf += ql.pack32(ql.os.uid)  # 28
        fstat64_buf += ql.pack32(ql.os.gid)  # 32
        fstat64_buf += ql.pack64(info.st_rdev)  # 40
        fstat64_buf += ql.pack64(0)  # 48
        fstat64_buf += ql.pack64(info.st_size)  # 56
        fstat64_buf += ql.pack32(info.st_blksize)  # 60
        fstat64_buf += ql.pack32(0)  # 64
        fstat64_buf += ql.pack64(info.st_blocks)  # 72
        fstat64_buf += ql.pack64(int(info.st_atime))  # 80
        fstat64_buf += ql.pack64(0)  # 88
        fstat64_buf += ql.pack64(int(info.st_mtime))  # 96
        fstat64_buf += ql.pack64(0)  # 104
        fstat64_buf += ql.pack64(int(info.st_ctime))  # 114
        fstat64_buf += ql.pack64(0)  # 120
    elif ql.archtype == QL_ARCH.MIPS:
        # struct stat is : a0 addr is : 0x7fffedc0
        # buf.st_dev offest 0 4 2049
        # buf.st_ino offest 10 8 2400362
        # buf.st_mode offest 18 4 16893
        # buf.st_nlink offest 1c 4 5
        # buf.st_uid offest 20 4 1000
        # buf.st_gid offest 24 4 1000
        # buf.st_rdev offest 28 4 0
        # buf.st_size offest 38 8 0
        # buf.st_blksize offest 58 4 4096
        # buf.st_blocks offest 60 8 136
        # buf.st_atime offest 40 4 1586616689
        # buf.st_mtime offest 48 4 1586616689
        # buf.st_ctime offest 50 4 1586616689
        if ql.platform == QL_OS.MACOS:
            fstat64_buf = ql.pack32s(info.st_dev)
        else:
            fstat64_buf = ql.pack32(info.st_dev)  # 4
        fstat64_buf += b'\x00' * 12  # 16
        fstat64_buf += ql.pack64(info.st_ino)  # 24
        fstat64_buf += ql.pack32(info.st_mode)
        fstat64_buf += ql.pack32(info.st_nlink)
        fstat64_buf += ql.pack32(ql.os.uid)
        fstat64_buf += ql.pack32(ql.os.gid)
        fstat64_buf += ql.pack32(info.st_rdev)
        fstat64_buf += b'\x00' * 12
        fstat64_buf += ql.pack64(info.st_size)
        fstat64_buf += ql.pack64(int(info.st_atime))
        fstat64_buf += ql.pack64(0)
        fstat64_buf += ql.pack64(int(info.st_mtime))
        fstat64_buf += ql.pack64(0)
        fstat64_buf += ql.pack64(int(info.st_ctime))
        fstat64_buf += ql.pack64(0)
        fstat64_buf += ql.pack32(info.st_blksize)
        fstat64_buf += ql.pack32(0)
        fstat64_buf += ql.pack64(info.st_blocks)
    elif ql.archtype == QL_ARCH.ARM:
        # pack fstatinfo
        if ql.platform == QL_OS.MACOS:
            fstat64_buf = ql.pack64s(info.st_dev)
        else:
            fstat64_buf = ql.pack64(info.st_dev)
        fstat64_buf += ql.pack32(0)
        fstat64_buf += ql.pack32(info.st_ino)
        fstat64_buf += ql.pack32(info.st_mode)
        fstat64_buf += ql.pack32(info.st_nlink)
        fstat64_buf += ql.pack32(info.st_uid)
        fstat64_buf += ql.pack32(info.st_gid)
        fstat64_buf += ql.pack64(info.st_rdev)  # ?? fstat_info.st_rdev
        fstat64_buf += ql.pack64(0)
        fstat64_buf += ql.pack64(info.st_size)
        fstat64_buf += ql.pack64(info.st_blksize)  # ?? fstat_info.st_blksize
        fstat64_buf += ql.pack64(info.st_blocks)  # ?? fstat_info.st_blocks
        fstat64_buf += ql.pack64(int(info.st_atime))
        fstat64_buf += ql.pack64(int(info.st_mtime))
        fstat64_buf += ql.pack64(int(info.st_ctime))
        fstat64_buf += ql.pack64(info.st_ino)

    else:
        # pack fstatinfo
        if ql.platform == QL_OS.MACOS:
            fstat64_buf = ql.pack64s(info.st_dev)
        else:
            fstat64_buf = ql.pack64(info.st_dev)
        fstat64_buf += ql.pack64(0x0000000300c30000)
        fstat64_buf += ql.pack32(info.st_mode)
        fstat64_buf += ql.pack32(info.st_nlink)
        fstat64_buf += ql.pack32(info.st_uid)
        fstat64_buf += ql.pack32(info.st_gid)
        fstat64_buf += ql.pack64(0x0000000000008800)  # ?? fstat_info.st_rdev
        fstat64_buf += ql.pack32(0xffffd257)
        fstat64_buf += ql.pack64(info.st_size)
        fstat64_buf += ql.pack32(0x00000400)  # ?? fstat_info.st_blksize
        fstat64_buf += ql.pack64(0x0000000000000000)  # ?? fstat_info.st_blocks
        fstat64_buf += ql.pack64(int(info.st_atime))
        fstat64_buf += ql.pack64(int(info.st_mtime))
        fstat64_buf += ql.pack64(int(info.st_ctime))
        fstat64_buf += ql.pack64(info.st_ino)
    return fstat64_buf


def statFamily(ql, path, ptr, name, stat_func, struct_func):
    file = (ql.mem.string(path))
    real_path = ql.os.transform_to_real_path(file)
    regreturn = 0
    try:
        info = stat_func(real_path)
    except OSError as e:
        ql.log.debug(f'{name}("{file}", {hex(ptr)}) read/write fail')
        return -e.errno
    else:
        buf = struct_func(ql, info)
        ql.mem.write(ptr, buf)
        ql.log.debug(f'{name}("{file}", {hex(ptr)}) write completed')
        return regreturn


def ql_syscall_chmod(ql, filename, mode, null1, null2, null3, null4):
    regreturn = 0
    filename = ql.mem.string(filename)
    ql.log.debug("chmod(%s,%d) = %d" % (filename, mode, regreturn))
    return regreturn


def ql_syscall_fstatat64(ql, fstatat64_fd, fstatat64_fname, fstatat64_buf, fstatat64_flag, *args, **kw):
    fstatat64_fname = ql.mem.string(fstatat64_fname)

    real_path = ql.os.transform_to_real_path(fstatat64_fname)
    relative_path = ql.os.transform_to_relative_path(fstatat64_fname)

    regreturn = -1
    if os.path.exists(real_path) == True:
        fstat64_info = Stat(real_path)

        # struct stat is : 80 addr is : 0x4000811bc8
        # buf.st_dev offest 0 8 0
        # buf.st_ino offest 8 8 0
        # buf.st_mode offest 10 4 0
        # buf.st_nlink offest 14 4 0
        # buf.st_uid offest 18 4 0
        # buf.st_gid offest 1c 4 0
        # buf.st_rdev offest 20 8 0
        # buf.st_size offest 30 8 274886889936
        # buf.st_blksize offest 38 4 8461328
        # buf.st_blocks offest 40 8 274877909532
        # buf.st_atime offest 48 8 274886368336
        # buf.st_mtime offest 58 8 274877909472
        # buf.st_ctime offest 68 8 274886368336
        # buf.__glibc_reserved offest 78 8
        fstat64_buf = ql.pack64(fstat64_info.st_dev)
        fstat64_buf += ql.pack64(fstat64_info.st_ino)
        fstat64_buf += ql.pack32(fstat64_info.st_mode)
        fstat64_buf += ql.pack32(fstat64_info.st_nlink)
        fstat64_buf += ql.pack32(1000)
        fstat64_buf += ql.pack32(1000)
        fstat64_buf += ql.pack64(fstat64_info.st_rdev)
        fstat64_buf += ql.pack64(0)
        fstat64_buf += ql.pack64(fstat64_info.st_size)
        fstat64_buf += ql.pack32(fstat64_info.st_blksize)
        fstat64_buf += ql.pack32(0)
        fstat64_buf += ql.pack64(fstat64_info.st_blocks)
        fstat64_buf += ql.pack64(int(fstat64_info.st_atime))
        fstat64_buf += ql.pack64(0)
        fstat64_buf += ql.pack64(int(fstat64_info.st_mtime))
        fstat64_buf += ql.pack64(0)
        fstat64_buf += ql.pack64(int(fstat64_info.st_ctime))
        fstat64_buf += ql.pack64(0)
        ql.mem.write(fstatat64_buf, fstat64_buf)
        regreturn = 0

    if regreturn == 0:
        ql.log.debug("Directory Found: %s" % relative_path)
    else:
        ql.log.debug("Directory Not Found: %s" % relative_path)

    return regreturn


def ql_syscall_fstat64(ql, fstat64_fd, fstat64_add, *args, **kw):

    if ql.os.fd[fstat64_fd].fstat() == -1:
        regreturn = 0

    elif fstat64_fd < 256 and ql.os.fd[fstat64_fd] != 0:
        user_fileno = fstat64_fd
        fstat64_info = ql.os.fd[user_fileno].fstat()
        fstat64_buf = create_stat64_struct(ql, fstat64_info)
        ql.mem.write(fstat64_add, fstat64_buf)
        regreturn = 0
    else:
        regreturn = -1

    if regreturn == 0:
        ql.log.debug("fstat64 write completed")
    else:
        ql.log.debug("fstat64 read/write fail")
    return regreturn


def ql_syscall_fstat(ql, fstat_fd, fstat_add, *args, **kw):
    if fstat_fd < 256 and ql.os.fd[fstat_fd] != 0 and hasattr(ql.os.fd[fstat_fd], "fstat"):
        user_fileno = fstat_fd
        fstat_info = ql.os.fd[user_fileno].fstat()
        fstat_buf = create_stat_struct(ql, fstat_info)
        ql.mem.write(fstat_add, fstat_buf)
        regreturn = 0
    else:
        regreturn = -1

    if regreturn == 0:
        ql.log.debug("fstat write completed")
    else:
        ql.log.debug("fstat read/write fail")
    return regreturn


# int stat(const char *path, struct stat *buf);
def ql_syscall_stat(ql, stat_path, stat_buf_ptr, *args, **kw):
    return statFamily(ql, stat_path, stat_buf_ptr, "stat", Stat, create_stat_struct)


# int stat64(const char *pathname, struct stat64 *buf);
def ql_syscall_stat64(ql, stat64_pathname, stat64_buf_ptr, *args, **kw):
    return statFamily(ql, stat64_pathname, stat64_buf_ptr, "stat64", Stat, create_stat64_struct)


def ql_syscall_lstat(ql, lstat_path, lstat_buf_ptr, *args, **kw):
    return statFamily(ql, lstat_path, lstat_buf_ptr, "lstat", Lstat, create_stat_struct)


def ql_syscall_lstat64(ql, lstat64_path, lstat64_buf_ptr, *args, **kw):
    return statFamily(ql, lstat64_path, lstat64_buf_ptr, "lstat64", Lstat, create_stat64_struct)


def ql_syscall_mknodat(ql, dirfd, pathname, mode, dev, *args, **kw):
    # fix me. dirfd(relative path) not implement.
    file_path = ql.mem.string(pathname)
    real_path = ql.os.transform_to_real_path(file_path)
    ql.log.debug("mknodat(%d, %s, 0%o, %d)" % (dirfd, real_path, mode, dev))
    try:
        os.mknod(real_path, mode, dev)
        regreturn = 0
    except:
        regreturn = -1
    return regreturn


def ql_syscall_mkdir(ql, pathname, mode, *args, **kw):
    file_path = ql.mem.string(pathname)
    real_path = ql.os.transform_to_real_path(file_path)
    ql.log.debug("mkdir(%s, 0%o)" % (real_path, mode))
    try:
        if not os.path.exists(real_path):
            os.mkdir(real_path, mode)
        regreturn = 0
    except:
        regreturn = -1
    return regreturn


def ql_syscall_umask(ql, mode, *args, **kw):
    oldmask = os.umask(mode)
    ql.log.debug("umask(0%o) return oldmask 0%o" % (mode, oldmask))
    regreturn = oldmask
    return regreturn
